"""File 19individualLevelParameters

:author: Michel Bierlaire, EPFL
:date: Wed Aug 26 14:56:49 2020

Calculation of the individual level parameters for model 05normalMixture
"""

import os
import pickle
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, DefineVariable, bioDraws, MonteCarlo

# Read the data
df = pd.read_csv('swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Here we use the "biogeme" way for backward compatibility
exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)

# Values of the parameters estimated by the model 05normalMixture
ASC_CAR = Beta('ASC_CAR', 0.137, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', -0.402, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', -1.28, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
B_TIME = Beta('B_TIME', -2.26, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1.65, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# Definition of new variables
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

# Definition of new variables: adding columns to the database
CAR_AV_SP = DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0), database)
TRAIN_AV_SP = DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0), database)
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0, database)
TRAIN_COST_SCALED = DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100, database
)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0, database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100, database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100, database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100, database)

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob_chosen = models.logit(V, av, CHOICE)

numerator = MonteCarlo(B_TIME_RND * prob_chosen)
denominator = MonteCarlo(prob_chosen)

simulate = {
    'Numerator': numerator,
    'Denominator': denominator,
    'Choice': CHOICE,
}

pickleFile = '19individualLevelParameters.pickle'
if os.path.isfile(pickleFile):
    with open(pickleFile, 'rb') as f:
        sim = pickle.load(f)
else:
    biosim = bio.BIOGEME(database, simulate, numberOfDraws=100000)
    sim = biosim.simulate()
    sim['Individual-level parameters'] = sim['Numerator'] / sim['Denominator']
    with open(pickleFile, 'wb') as f:
        pickle.dump(sim, f)



"""File 07discreteMixture.py

:author: Michel Bierlaire, EPFL
:date: Sun Sep  8 00:06:20 2019

 Example of a discrete mixture of logit (or latent class model)
 Three alternatives: Train, Car and Swissmetro
 SP data
"""

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
import biogeme.messaging as msg
from biogeme.expressions import Beta, DefineVariable, log

# Read the data
df = pd.read_csv('swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
# print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
# remove = (((database.data.PURPOSE != 1) &
#           (database.data.PURPOSE != 3)) |
#          (database.data.CHOICE == 0))
# database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

PROB_CLASS1 = Beta('PROB_CLASS1', 0.5, 0, 1, 0)
PROB_CLASS2 = 1 - PROB_CLASS1

# Definition of new variables
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

# Definition of new variables: adding columns to the database
CAR_AV_SP = DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0), database)
TRAIN_AV_SP = DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0), database)
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0, database)
TRAIN_COST_SCALED = DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100, database
)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0, database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100, database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100, database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100, database)

# Definition of the utility functions for latent class 1, where the
# time coefficient is zero
V11 = ASC_TRAIN + B_COST * TRAIN_COST_SCALED
V12 = ASC_SM + B_COST * SM_COST_SCALED
V13 = ASC_CAR + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V1 = {1: V11, 2: V12, 3: V13}

# Definition of the utility functions for latent class 2, whete the
# time coefficient is estimated
V21 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V22 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V23 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V2 = {1: V21, 2: V22, 3: V23}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}


# The choice model is a discrete mixture of logit, with availability conditions
prob1 = models.logit(V1, av, CHOICE)
prob2 = models.logit(V2, av, CHOICE)
prob = PROB_CLASS1 * prob1 + PROB_CLASS2 * prob2
logprob = log(prob)

# Define level of verbosity
logger = msg.bioMessage()
logger.setSilent()
# logger.setWarning()
# logger.setGeneral()
# logger.setDetailed()

# Create the Biogeme object
biogeme = bio.BIOGEME(database, logprob)
biogeme.modelName = '07discreteMixture'

# Estimate the parameters
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)



"""Implementation of the main Biogeme class that combines the database
and the model specification.

:author: Michel Bierlaire
:date: Tue Mar 26 16:45:15 2019

"""

# Too constraining
# pylint: disable=invalid-name,
# pylint: disable=too-many-arguments, too-many-locals,
# pylint: disable=too-many-statements, too-many-branches,
# pylint: disable=too-many-instance-attributes, too-many-lines,
# pylint: disable=too-many-function-args, invalid-unary-operand-type


import multiprocessing as mp
from datetime import datetime
import pickle
import numpy as np
import pandas as pd
import tqdm

import biogeme.database as db
import biogeme.cbiogeme as cb
import biogeme.expressions as eb
import biogeme.results as res
import biogeme.exceptions as excep
import biogeme.filenames as bf
import biogeme.messaging as msg
import biogeme.optimization as opt
from biogeme import tools
from biogeme.algorithms import functionToMinimize


# import yep


class BIOGEME:
    """Main class that combines the database and the model specification.

    It works in two modes: estimation and simulation.

    """

    def __init__(
        self,
        database,
        formulas,
        userNotes=None,
        numberOfThreads=None,
        numberOfDraws=1000,
        seed=None,
        skipAudit=False,
        removeUnusedVariables=True,
        displayUsedVariables=False,
        suggestScales=True,
        missingData=99999,
    ):
        """Constructor

        :param database: choice data.
        :type database: :class:`biogeme.database.Database`

        :param formulas: expression or dictionary of expressions that
             define the model specification.  The concept is that each
             expression is applied to each entry of the database. The
             keys of the dictionary allow to provide a name to each
             formula.  In the estimation mode, two formulas are
             needed, with the keys 'loglike' and 'weight'. If only one
             formula is provided, it is associated with the label
             'loglike'. If no formula is labeled 'weight', the weight
             of each piece of data is supposed to be 1.0. In the
             simulation mode, the labels of each formula are used as
             labels of the resulting database.
        :type formulas: :class:`biogeme.expressions.Expression`, or
                        dict(:class:`biogeme.expressions.Expression`)

        :param userNotes: these notes will be included in the report file.
        :type userNotes: str

        :param numberOfThreads: multi-threading can be used for
            estimation. This parameter defines the number of threads
            to be used. If the parameter is set to None, the number of
            available threads is calculated using
            cpu_count(). Ignored in simulation mode. Defaults: None.
        :type numberOfThreads:  int

        :param numberOfDraws: number of draws used for Monte-Carlo
            integration. Default: 1000.
        :type numberOfDraws: int

        :param seed: seed used for the pseudo-random number
            generation. It is useful only when each run should
            generate the exact same result. If None, a new seed is
            used at each run. Default: None.
        :type seed: int

        :param skipAudit: if True, does not check the validity of the
            formulas. It may save significant amount of time for large
            models and large data sets. Default: False.
        :type skipAudit: bool

        :param removeUnusedVariables: if True, all variables not used
            in the expression are removed from the database. Default:
            True.
        :type removeUnusedVariables: bool

        :param displayUsedVariables: if True, displays all the
            variables used in the formulas. Default: False.
        :type displayUsedVariables: bool

        :param suggestScales: if True, Biogeme suggests the scaling of
            the variables in the database. Default: True.
            See also :func:`biogeme.database.Database.suggestScaling`

        :type suggestScales: bool.

        :param missingData: if one variable has this value, it is
           assumed that a data is missing and an exception will be
           triggered. Default: 99999.
        :type missingData: float

        :raise biogemeError: an audit of the formulas is performed.
           If a formula has issues, an error is detected and an
           exception is raised.

        """

        self.logger = msg.bioMessage()
        """Logger that controls the output of
        messages to the screen and log file.
        Type: class :class:`biogeme.messaging.bioMessage`."""

        if not skipAudit:
            database.data = database.data.replace({True: 1, False: 0})
            listOfErrors, listOfWarnings = database._audit()
            if listOfWarnings:
                self.logger.warning('\n'.join(listOfWarnings))
            if listOfErrors:
                self.logger.warning('\n'.join(listOfErrors))
                raise excep.biogemeError('\n'.join(listOfErrors))

        self.loglikeName = 'loglike'
        """ Keyword used for the name of the loglikelihood formula.
        Default: 'loglike'"""


        self.weightName = 'weight'
        """Keyword used for the name of the weight formula. Default: 'weight'
        """

        self.modelName = 'biogemeModelDefaultName'
        """Name of the model. Default: 'biogemeModelDefaultName'
        """
        self.monteCarlo = False
        """ ``monteCarlo`` is True if one of the expressions involves a
        Monte-Carlo integration.
        """

        if seed is not None:
            np.random.seed(seed)

        self.saveIterations = True
        """If True, the current iterate is saved after each iteration, in a
        file named ``__[modelName].iter``, where ``[modelName]`` is the
        name given to the model. If such a file exists, the starting
        values for the estimation are replaced by the values saved in
        the file.
        """

        if not isinstance(formulas, dict):
            if not isinstance(formulas, eb.Expression):
                raise excep.biogemeError(
                    f'Expression {formulas} is not of type '
                    f'biogeme.expressions.Expression. '
                    f'It is of type {type(formulas)}'
                )

            self.loglike = formulas
            """ Object of type :class:`biogeme.expressions.Expression`
            calculating the formula for the loglikelihood
            """

            self.weight = None
            """ Object of type :class:`biogeme.expressions.Expression`
            calculating the weight of each observation in the
            sample.
            """

            self.formulas = dict({self.loglikeName: formulas})
            """ Dictionary containing Biogeme formulas of type
            :class:`biogeme.expressions.Expression`.
            The keys are the names of the formulas.
            """
        else:
            # Verify the validity of the formulas
            for k, f in formulas.items():
                if not isinstance(f, eb.Expression):
                    raise excep.biogemeError(
                        f'Expression for "{k}" is not of type '
                        f'biogeme.expressions.Expression. '
                        f'It is of type {type(f)}'
                    )
            self.loglike = formulas.get(self.loglikeName)
            self.weight = formulas.get(self.weightName)
            self.formulas = formulas

        self.database = database #: :class:`biogeme.database.Database` object

        self.userNotes = userNotes #: User notes

        self.missingData = missingData #: code for missing data


        self.lastSample = None
        """ keeps track of the sample of data used to calculate the
        stochastic gradient / hessian
        """

        self.initLogLike = None #: Init value of the likelihood function

        self.nullLogLike = None #: Log likelihood of the null model

        self.usedVariables = set() #: set of variables used in the formulas.
        for k, f in self.formulas.items():
            myvars = f.setOfVariables()
            missingVariables = [
                v for v in myvars if v not in self.database.data
            ]
            if missingVariables:
                errorMsg = (
                    f'Variables in formula {k} missing in the database: '
                    f'{missingVariables}'
                )
                raise excep.biogemeError(errorMsg)
            self.usedVariables |= f.setOfVariables()
        if self.database.isPanel():
            self.usedVariables.add(self.database.panelColumn)
        if displayUsedVariables:
            self.logger.general(
                f'List of used variables: {self.usedVariables}'
            )
        if removeUnusedVariables:
            unusedVariables = (
                set(self.database.data.columns) - self.usedVariables
            )
            error_msg = (
                f'Remove {len(unusedVariables)} '
                'unused variables from the database '
                f'as only {len(self.usedVariables)} are used.'
            )
            self.logger.general(error_msg)
            self.database.data = self.database.data.drop(
                columns=list(unusedVariables)
            )

        if suggestScales:
            suggestedScales = self.database.suggestScaling(
                columns=self.usedVariables
            )
            if not suggestedScales.empty:
                self.logger.detailed(
                    'It is suggested to scale the following variables.'
                )
                for _, row in suggestedScales.iterrows():
                    error_msg = (
                        f'Multiply {row["Column"]} by\t{row["Scale"]} '
                        'because the largest (abs) value is\t'
                        f'{row["Largest"]}'
                    )
                    self.logger.detailed(error_msg)
                error_msg = (
                    'To remove this feature, set the parameter '
                    'suggestScales to False when creating the '
                    'BIOGEME object.'
                )
                self.logger.detailed(error_msg)

        if not skipAudit:
            self._audit()

        self._prepareDatabaseForFormula()
        self._prepareLiterals()
        self.theC = cb.pyBiogeme(len(self.freeBetaNames))
        if self.database.isPanel():
            self.theC.setPanel(True)
            self.theC.setDataMap(self.database.individualMap)
        # Transfer the data to the C++ formula
        self.theC.setData(self.database.data)
        self.theC.setMissingData(self.missingData)

        self.generateHtml = True
        """ Boolean variable, True if the HTML file with the results must
        be generated.
        """

        self.generatePickle = True
        """ Boolean variable, True if the pickle file with the results must
        be generated.
        """

        self.columnForBatchSamplingWeights = None
        """ Name of the column defining weights for batch sampling in
        stochastic optimization.
        """

        self.numberOfThreads = (
            mp.cpu_count() if numberOfThreads is None else numberOfThreads
        )
        """ Number of threads used for parallel computing. Default: the number
        of available CPU.
        """

        start_time = datetime.now()
        self._generateDraws(numberOfDraws)
        if self.monteCarlo:
            self.theC.setDraws(self.database.theDraws)
        self.drawsProcessingTime = datetime.now() - start_time
        """ Time needed to generate the draws. """

        if self.loglike is not None:

            self.loglikeSignatures = self.loglike.getSignature()
            """ Internal signature of the formula for the loglikelihood."""
            if self.weight is None:
                self.theC.setExpressions(
                    self.loglikeSignatures, self.numberOfThreads
                )
            else:
                self.weightSignatures = self.weight.getSignature()
                """ Internal signature of the formula for the weight."""
                self.theC.setExpressions(
                    self.loglikeSignatures,
                    self.numberOfThreads,
                    self.weightSignatures,
                )

        self.bootstrap_time = None
        """ Time needed to calculate the bootstrap standard errors"""


        self.bootstrap_results = None #: Results of the bootstrap calculation.

        self.optimizationMessages = None
        """ Information provided by the optimization algorithm
        after completion.
        """

        self.algoParameters = None
        """ Parameters to be transferred to the optimization algorithm
        """


        self.algorithm = None #: Optimization algorithm

        self.bestIteration = None #: Store the best iteration found so far.


    def _saveIterationsFileName(self):
        """
        :return: The name of the file where the iterations are saved.
        :rtype: str
        """
        return f'__{self.modelName}.iter'

    def _audit(self):
        """Each expression provides an audit function, that verifies its
        validity. Each formula is audited, and the list of errors
        and warnings reported.

        :raise biogemeError: if the formula has issues, an error is
                             detected and an exception is raised.

        """

        listOfErrors = []
        listOfWarnings = []
        for v in self.formulas.values():
            err, war = v.audit(self.database)
            listOfErrors += err
            listOfWarnings += war
        if listOfWarnings:
            self.logger.warning('\n'.join(listOfWarnings))
        if listOfErrors:
            self.logger.warning('\n'.join(listOfErrors))
            raise excep.biogemeError('\n'.join(listOfErrors))

    def _generateDraws(self, numberOfDraws):
        """If Monte-Carlo integration is involved in one of the formulas, this
           function instructs the database to generate the draws.

        Args:
            numberOfDraws: self explanatory (int)
        """

        # Number of draws for Monte-Carlo integration.
        self.numberOfDraws = numberOfDraws
        # Draws
        self.monteCarlo = len(self.allDraws) > 0
        if self.monteCarlo:
            self.database.generateDraws(
                self.allDraws, self.drawNames, numberOfDraws
            )

    def _prepareDatabaseForFormula(self, sample=None):
        # Prepare the dataset.
        if sample is None:
            if self.lastSample == 1.0:
                # We continue to use the full data set. Nothing to be done.
                return
            self.lastSample = 1.0
            self.database.useFullSample()
        else:
            # Check if the sample size is valid
            if sample <= 0 or sample > 1.0:
                error_msg = (
                    f'The value of the parameter sample must be '
                    f'strictly between 0.0 and 1.0,'
                    f' and not {sample}'
                )
                raise ValueError(error_msg)

            if sample == 1.0:
                self.database.useFullSample()
            else:
                self.logger.detailed(f'Use {100*sample}% of the data.')
                if self.database.isPanel():
                    self.database.sampleIndividualMapWithoutReplacement(
                        sample, self.columnForBatchSamplingWeights
                    )
                else:
                    self.database.sampleWithoutReplacement(
                        sample, self.columnForBatchSamplingWeights
                    )
            self.lastSample = sample

        # Rebuild the map for panel data
        if self.database.isPanel():
            self.database.buildPanelMap()

    def getBoundsOnBeta(self, betaName):
        """Returns the bounds on the parameter as defined by the user.

        :param betaName: name of the parameter
        :type betaName: string
        :return: lower bound, upper bound
        :rtype: tuple
        :raises biogemeError: if the name of the parameter is not found.
        """

        if betaName not in self.freeBetaNames:
            raise excep.biogemeError(f'Unknown parameter {betaName}')
        index = self.freeBetaNames.index(betaName)
        return self.bounds[index]

    def _prepareLiterals(self):
        """Extract from the formulas the literals (parameters,
        variables, random variables) and decide a numbering convention.
        """

        collectionOfFormulas = [f for k, f in self.formulas.items()]
        variableNames = list(self.database.data.columns.values)

        (
            self.elementaryExpressionIndex,
            self.allFreeBetas,
            self.freeBetaNames,
            self.allFixedBetas,
            self.fixedBetaNames,
            self.allRandomVariables,
            self.randomVariableNames,
            self.allDraws,
            self.drawNames,
        ) = eb.defineNumberingOfElementaryExpressions(
            collectionOfFormulas, variableNames
        )

        # List of tuples (ell, u) containing the lower and upper bounds
        # for each free parameter
        self.bounds = list()
        for x in self.freeBetaNames:
            self.bounds.append(
                (self.allFreeBetas[x].lb, self.allFreeBetas[x].ub)
            )
        # List of ids of the free beta parameters (those to be estimated)
        self.betaIds = list(range(len(self.freeBetaNames)))

        # List of initial values of the free beta parameters (those to be
        # estimated)
        self.betaInitValues = [
            float(self.allFreeBetas[x].initValue) for x in self.freeBetaNames
        ]
        # Values of the fixed parameters (not estimated).
        self.fixedBetaValues = [
            float(self.allFixedBetas[x].initValue) for x in self.fixedBetaNames
        ]

    def calculateNullLoglikelihood(self, avail):
        """Calculate the log likelihood of the null model that predicts equal
        probability for each alternative

        :param avail: list of expressions to evaluate the
                      availability conditions for each alternative.
        :type avail: list of :class:`biogeme.expressions.Expression`

        :return: value of the log likelihood
        :rtype: float

        """
        expression = -eb.log(eb.bioMultSum(avail))
        self.nullLogLike = self.database.sumFromDatabase(expression)
        return self.nullLogLike

    def calculateInitLikelihood(self):
        """Calculate the value of the log likelihood function

        The default values of the parameters are used.

        :return: value of the log likelihood.
        :rtype: float.
        """

        # Value of the loglikelihood for the default values of the parameters.
        self.initLogLike = self.calculateLikelihood(
            self.betaInitValues, scaled=False
        )
        return self.initLogLike

    def calculateLikelihood(self, x, scaled, batch=None):
        """Calculates the value of the log likelihood function

        :param x: vector of values for the parameters.
        :type x: list(float)

        :param scaled: if True, the value is divided by the number of
                       observations used to calculate it. In this
                       case, the values with different sample sizes
                       are comparable. Default: True
        :type scaled: bool

        :param batch: if not None, calculates the likelihood on a
                       random sample of the data. The value of the
                       parameter must be strictly between 0 and 1, and
                       represents the share of the data that will be
                       used. Default: None
        :type batch: float

        :return: the calculated value of the log likelihood
        :rtype: float.

        :raises ValueError: if the length of the list x is incorrect.

        """

        if len(x) != len(self.betaInitValues):
            error_msg = (
                f'Input vector must be of length '
                f'{len(self.betaInitValues)} and '
                f'not {len(x)}'
            )
            raise ValueError(error_msg)

        self._prepareDatabaseForFormula(batch)
        f = self.theC.calculateLikelihood(x, self.fixedBetaValues)

        self.logger.detailed(
            f'Log likelihood (N = {self.database.getSampleSize()}): {f:10.7g}'
        )

        if scaled:
            return f / float(self.database.getSampleSize())

        return f

    def calculateLikelihoodAndDerivatives(
        self, x, scaled, hessian=False, bhhh=False, batch=None
    ):
        """Calculate the value of the log likelihood function
        and its derivatives.

        :param x: vector of values for the parameters.
        :type x: list(float)

        :param scaled: if True, the results are devided by the number of
            observations.
        :type scaled: bool

        :param hessian: if True, the hessian is calculated. Default: False.
        :type hessian: bool

        :param bhhh: if True, the BHHH matrix is calculated. Default: False.
        :type bhhh: bool

        :param batch: if not None, calculates the likelihood on a
                       random sample of the data. The value of the
                       parameter must be strictly between 0 and 1, and
                       represents the share of the data that will be
                       used. Default: None
        :type batch: float


        :return: f, g, h, bh where

                - f is the value of the function (float)
                - g is the gradient (numpy.array)
                - h is the hessian (numpy.array)
                - bh is the BHHH matrix (numpy.array)

        :rtype: tuple  float, numpy.array, numpy.array, numpy.array

        :raises ValueError: if the length of the list x is incorrect

        :raises biogemeError: if the norm of the gradient is not finite, an
            error is raised.
        """

        n = len(x)
        if n != len(self.betaInitValues):
            error_msg = (
                f'Input vector must be of length '
                f'{len(self.betaInitValues)} and not {len(x)}'
            )
            raise ValueError(error_msg)
        self._prepareDatabaseForFormula(batch)

        g = np.empty(n)
        h = np.empty([n, n])
        bh = np.empty([n, n])

        f, g, h, bh = self.theC.calculateLikelihoodAndDerivatives(
            x, self.fixedBetaValues, self.betaIds, g, h, bh, hessian, bhhh
        )

        hmsg = ''
        if hessian:
            hmsg = f'Hessian norm:  {np.linalg.norm(h):10.1g}'
        bhhhmsg = ''
        if bhhh:
            bhhhmsg = f'BHHH norm:  {np.linalg.norm(bh):10.1g}'
        gradnorm = np.linalg.norm(g)
        self.logger.general(
            f'Log likelihood (N = {self.database.getSampleSize()}): {f:10.7g}'
            f' Gradient norm: {gradnorm:10.1g}'
            f' {hmsg} {bhhhmsg}'
        )

        if not np.isfinite(gradnorm):
            error_msg = f'The norm of the gradient is {gradnorm}: g={g}'
            raise excep.biogemeError(error_msg)

        if self.saveIterations:
            if self.bestIteration is None:
                self.bestIteration = f
            if f >= self.bestIteration:
                with open(self._saveIterationsFileName(), 'w') as pf:
                    for i, v in enumerate(x):
                        print(f'{self.freeBetaNames[i]} = {v}', file=pf)

        if scaled:
            N = float(self.database.getSampleSize())
            if N == 0:
                raise excep.biogemeError(f'Sample size is {N}')

            return (
                f / N,
                np.asarray(g) / N,
                np.asarray(h) / N,
                np.asarray(bh) / N,
            )
        return f, np.asarray(g), np.asarray(h), np.asarray(bh)

    def likelihoodFiniteDifferenceHessian(self, x):
        """Calculate the hessian of the log likelihood function using finite
        differences.

        May be useful when the analytical hessian has numerical issues.

        :param x: vector of values for the parameters.
        :type x: list(float)

        :return: finite differences approximation of the hessian.
        :rtype: numpy.array

        :raises ValueError: if the length of the list x is incorrect

        """

        def theFunction(x):
            f, g, _, _ = self.calculateLikelihoodAndDerivatives(
                x, scaled=False, hessian=False, bhhh=False
            )
            return f, np.asarray(g)

        return tools.findiff_H(theFunction, np.asarray(x))

    def checkDerivatives(self, verbose=False):
        """Verifies the implementation of the derivatives.

        It compares the analytical version with the finite differences
        approximation.

        :param verbose: if True, the comparisons are reported. Default: False.
        :type verbose: bool

        :rtype: tuple.

        :return: f, g, h, gdiff, hdiff where

            - f is the value of the function,
            - g is the analytical gradient,
            - h is the analytical hessian,
            - gdiff is the difference between the analytical and the
              finite differences gradient,
            - hdiff is the difference between the analytical and the
              finite differences hessian,

        """

        def theFunction(x):
            """Wrapper function to use tools.checkDerivatives"""
            f, g, h, _ = self.calculateLikelihoodAndDerivatives(
                x, scaled=False, hessian=True, bhhh=False
            )
            return f, np.asarray(g), np.asarray(h)

        return tools.checkDerivatives(
            theFunction,
            np.asarray(self.betaInitValues),
            self.freeBetaNames,
            verbose,
        )

    def loadSavedIteration(self, filename='__savedIterations.txt'):
        """
        Obsolete function
        """
        message = (
            'The function loadSavedIterations is obsolete. It is '
            'sufficient to set the parameter saveIterations to True '
            'or False to control the process. Therefore, there '
            'is no need to call the function loadSavedIteration anymore.'
        )
        raise excep.biogemeError(message)

    def _loadSavedIteration(self):
        """Reads the values of the parameters from a text file where each line
        has the form name_of_beta = value_of_beta, and use these values in all
        formulas.

        """
        filename = self._saveIterationsFileName()
        betas = {}
        try:
            with open(filename) as fp:
                for line in fp:
                    ell = line.split('=')
                    betas[ell[0].strip()] = float(ell[1])
            self.changeInitValues(betas)
            self.logger.detailed(f'Parameter values restored from {filename}')
        except IOError:
            self.logger.warning(
                f'Cannot read file {filename}. Statement is ignored.'
            )

    def setRandomInitValues(self, defaultBound=100.0):
        """Modifies the initial values of the parameters in all formulas,
        using randomly generated values. The value is drawn from a
        uniform distribution on the interval defined by the
        bounds.

        :param defaultBound: If the upper bound is missing, it is
            replaced by this value. If the lower bound is missing, it is
            replaced by the opposite of this value. Default: 100.
        :type defaultBound: float
        """
        randomBetas = {
            name: np.random.uniform(
                low=-defaultBound if beta.lb is None else beta.lb,
                high=defaultBound if beta.ub is None else beta.ub,
            )
            for name, beta in self.allFreeBetas.items()
        }
        self.changeInitValues(randomBetas)

    def changeInitValues(self, betas):
        """Modifies the initial values of the pameters in all formula

        :param betas: dictionary where the keys are the names of the
                      parameters, and the values are the new value for
                      the parameters.
        :type betas: dict(string:float)
        """
        if self.loglike is not None:
            self.loglike.changeInitValues(betas)
        if self.weight is not None:
            self.weight.changeInitValues(betas)
        for _, f in self.formulas.items():
            f.changeInitValues(betas)
        for i, name in enumerate(self.freeBetaNames):
            value = betas.get(name)
            if value is not None:
                self.betaInitValues[i] = value

    def estimate(
        self,
        bootstrap=0,
        algorithm=opt.simpleBoundsNewtonAlgorithmForBiogeme,
        algoParameters=None,
    ):

        """Estimate the parameters of the model.

        :param bootstrap: number of bootstrap resampling used to
               calculate the variance-covariance matrix using
               bootstrapping. If the number is 0, bootstrapping is not
               applied. Default: 0.
        :type bootstrap: int

        :param algorithm: optimization algorithm to use for the
               maximum likelihood estimation. Default: Biogeme's
               Newton's algorithm with simple bounds.
        :type algorithm: function

        :param algoParameters: parameters to transfer to the optimization
            algorithm
        :type algoParameters: dict

        :return: object containing the estimation results.
        :rtype: biogeme.bioResults

        Example::

            # Create an instance of biogeme
            biogeme  = bio.BIOGEME(database, logprob)

            # Gives a name to the model
            biogeme.modelName = 'mymodel'

            # Estimate the parameters
            results = biogeme.estimate()

        :raises biogemeError: if no expression has been provided for the
            likelihood
        """

        if self.loglike is None:
            raise excep.biogemeError(
                'No log likelihood function has been specificed'
            )
        if len(self.freeBetaNames) == 0:
            raise excep.biogemeError(
                f'There is no parameter to estimate'
                f' in the formula: {self.loglike}.'
            )

        if self.saveIterations:
            self.logger.general(
                f'*** Initial values of the parameters are '
                f'obtained from the file {self._saveIterationsFileName()}'
            )
            self._loadSavedIteration()
        self.algorithm = algorithm
        self.algoParameters = algoParameters

        self.calculateInitLikelihood()
        self.bestIteration = None

        start_time = datetime.now()
        #        yep.start('profile.out')

        #        yep.stop()

        output = self.optimize(self.betaInitValues)
        xstar, optimizationMessages = output
        # Running time of the optimization algorithm
        optimizationMessages['Optimization time'] = datetime.now() - start_time
        # Information provided by the optimization algorithm after completion.
        self.optimizationMessages = optimizationMessages

        fgHb = self.calculateLikelihoodAndDerivatives(
            xstar, scaled=False, hessian=True, bhhh=True
        )
        if not np.isfinite(fgHb[2]).all():
            warning_msg = (
                'Numerical problems in calculating '
                'the analytical hessian. Finite differences'
                ' is tried instead.'
            )
            self.logger.warning(warning_msg)
            finDiffHessian = self.likelihoodFiniteDifferenceHessian(xstar)
            if not np.isfinite(fgHb[2]).all():
                self.logger.warning(
                    'Numerical problems with finite '
                    'difference hessian as well.'
                )
            else:
                fgHb = fgHb[0], fgHb[1], finDiffHessian, fgHb[3]
        # numpy array, of size B x K,
        # where
        #        - B is the number of bootstrap iterations
        #        - K is the number pf parameters to estimate
        self.bootstrap_results = None
        if bootstrap > 0:
            start_time = datetime.now()

            self.logger.general(
                f'Re-estimate the model {bootstrap} times for bootstrapping'
            )
            self.bootstrap_results = np.empty(shape=[bootstrap, len(xstar)])
            hideProgress = self.logger.screenLevel == 0
            self.logger.temporarySilence()
            for b in tqdm.tqdm(range(bootstrap), disable=hideProgress):
                if self.database.isPanel():
                    sample = self.database.sampleIndividualMapWithReplacement()
                    self.theC.setDataMap(sample)
                else:
                    sample = self.database.sampleWithReplacement()
                    self.theC.setData(sample)
                x_br, _ = self.optimize(xstar)
                self.bootstrap_results[b] = x_br

            # Time needed to generate the bootstrap results
            self.bootstrap_time = datetime.now() - start_time
            self.logger.resume()
        rawResults = res.rawResults(
            self, xstar, fgHb, bootstrap=self.bootstrap_results
        )
        r = res.bioResults(rawResults)
        if self.generateHtml:
            r.writeHtml()
        if self.generatePickle:
            r.writePickle()
        return r

    def quickEstimate(
        self,
        algorithm=opt.simpleBoundsNewtonAlgorithmForBiogeme,
        algoParameters=None,
    ):

        """| Estimate the parameters of the model. Same as estimate, where any
             extra calculation is skipped (init loglikelihood,
             t-statistics, etc.)

        :param algorithm: optimization algorithm to use for the
               maximum likelihood estimation.Default: Biogeme's
               Newton's algorithm with simple bounds.
        :type algorithm: function

        :param algoParameters: parameters to transfer to the optimization
            algorithm
        :type algoParameters: dict

        :return: object containing the estimation results.
        :rtype: biogeme.results.bioResults

        Example::

            # Create an instance of biogeme
            biogeme  = bio.BIOGEME(database, logprob)

            # Gives a name to the model
            biogeme.modelName = 'mymodel'

            # Estimate the parameters
            results = biogeme.quickEstimate()

        :raises biogemeError: if no expression has been provided for the
            likelihood

        """

        if self.loglike is None:
            raise excep.biogemeError(
                'No log likelihood function has been specificed'
            )
        if len(self.freeBetaNames) == 0:
            raise excep.biogemeError(
                f'There is no parameter to estimate'
                f' in the formula: {self.loglike}.'
            )

        self.algorithm = algorithm
        self.algoParameters = algoParameters

        start_time = datetime.now()
        #        yep.start('profile.out')

        #        yep.stop()

        output = self.optimize(self.betaInitValues)
        xstar, optimizationMessages = output
        # Running time of the optimization algorithm
        optimizationMessages['Optimization time'] = datetime.now() - start_time
        # Information provided by the optimization algorithm after completion.
        self.optimizationMessages = optimizationMessages

        f = self.calculateLikelihood(xstar, scaled=False)

        fgHb = f, None, None, None
        rawResults = res.rawResults(
            self, xstar, fgHb, bootstrap=self.bootstrap_results
        )
        r = res.bioResults(rawResults)
        return r

    def validate(self, estimationResults, validationData):
        """Perform out-of-sample validation.

        The function performs the following tasks:

          - each slice defines a validation set (the slice itself)
            and an estimation set (the rest of the data),
          - the model is re-estimated on the estimation set,
          - the estimated model is applied on the validation set,
          - the value of the log likelihood for each observation is reported.

        :param estimationResults: results of the model estimation based on the
            full data.
        :type estimationResults: biogeme.results.bioResults

        :param validationData: list of estimation and validation data sets
        :type validationData: list(tuple(pandas.DataFrame, pandas.DataFrame))

        :return: a list containing as many items as slices. Each item
                 is the result of the simulation on the validation set.
        :rtype: list(pandas.DataFrame)

        :raises biogemeError: An error is raised if the database is structured
            as panel data.
        """
        if self.database.isPanel():
            raise excep.biogemeError(
                'Validation for panel data is not yet implemented'
            )

        keepDatabase = self.database

        allSimulationResults = []
        for v in validationData:
            # v[0] is the estimation data set
            self.database = db.Database('Estimation data', v[0])
            self.loglike.changeInitValues(estimationResults.getBetaValues())
            results = self.estimate()
            simulate = {'Loglikelihood': self.loglike}
            simBiogeme = BIOGEME(
                db.Database('Validation data', v[1]), simulate
            )
            simResult = simBiogeme.simulate(results.getBetaValues())
            allSimulationResults.append(simResult)
        self.database = keepDatabase
        if self.generatePickle:
            fname = f'{self.modelName}_validation'
            pickleFileName = bf.getNewFileName(fname, 'pickle')
            with open(pickleFileName, 'wb') as f:
                pickle.dump(allSimulationResults, f)
            self.logger.general(
                f'Simulation results saved in file {pickleFileName}'
            )

        return allSimulationResults

    def optimize(self, startingValues=None):
        """Calls the optimization algorithm. The function self.algorithm
        is called.

        :param startingValues: starting point for the algorithm
        :type startingValues: list(float)

        :return: x, messages

           - x is the solution generated by the algorithm,
           - messages is a dictionary describing several information about the
             algorithm

        :rtype: numpay.array, dict(str:object)


        :raises biogemeError: an error is raised if no algorithm is specified.
        """

        theFunction = negLikelihood(
            like=self.calculateLikelihood,
            like_deriv=self.calculateLikelihoodAndDerivatives,
            scaled=True,
        )

        if startingValues is None:
            startingValues = self.betaInitValues

        if self.algorithm is None:
            err = (
                'An algorithm must be specified. The CFSQP algorithm '
                'is not available anymore.'
            )
            raise excep.biogemeError(err)

        results = self.algorithm(
            theFunction, startingValues, self.bounds, self.algoParameters
        )
        return results

    def simulate(self, theBetaValues=None):
        """Applies the formulas to each row of the database.

        :param theBetaValues: values of the parameters to be used in
                the calculations. If None, the default values are
                used. Default: None.
        :type theBetaValues: dict(str, float)

        :return: a pandas data frame with the simulated value. Each
              row corresponds to a row in the database, and each
              column to a formula.

        :rtype: Pandas data frame

        Example::

              # Read the estimation results from a file
              results = res.bioResults(pickleFile = 'myModel.pickle')
              # Simulate the formulas using the nominal values
              simulatedValues = biogeme.simulate(betaValues)

        :raises biogemeError: if the number of parameters is incorrect

        """

        if self.database.isPanel():
            error_msg = (
                'Simulation for panel data is not yet'
                ' implemented. Remove the "panel" '
                'statement to simulate each observation.'
            )
            raise excep.biogemeError(error_msg)

        if theBetaValues is None:
            betaValues = self.betaInitValues
        else:
            if not isinstance(theBetaValues, dict):
                err = (
                    'Deprecated. A dictionary must be provided. '
                    'It can be obtained from results.getBetaValues()'
                )
                raise excep.biogemeError(err)
            for x in theBetaValues.keys():
                if x not in self.freeBetaNames:
                    self.logger.warning(f'Parameter {x} not present in the model')
            betaValues = list()
            for i, x in enumerate(self.freeBetaNames):
                if x in theBetaValues:
                    betaValues.append(theBetaValues[x])
                else:
                    self.logger.warning(
                        f'Simulation: initial value of {x} not provided.'
                    )
                    betaValues.append(self.betaInitValues[i])

        output = pd.DataFrame(index=self.database.data.index)
        formulas = [v.getSignature() for v in self.formulas.values()]
        result = self.theC.simulateSeveralFormulas(
            formulas,
            betaValues,
            self.fixedBetaValues,
            self.database.data,
            self.numberOfThreads,
        )
        for key, r in zip(self.formulas.keys(), result):
            output[key] = r
        return output

    def oldsimulate(self, theBetaValues=None):
        """Applies the formulas to each row of the database. This is the old
        implementation. To be removed in future versions.

        :param theBetaValues: values of the parameters to be used in
                the calculations. If None, the default values are
                used. Default: None.
        :type theBetaValues: dict(str, float)

        :return: a pandas data frame with the simulated value. Each
              row corresponds to a row in the database, and each
              column to a formula.

        :rtype: Pandas data frame

        :raises biogemeError: if the number of parameters is incorrect

        """

        if self.database.isPanel():
            error_msg = (
                'Simulation for panel data is not yet'
                ' implemented. Remove the "panel" '
                'statement to simulate each observation.'
            )
            raise excep.biogemeError(error_msg)

        if theBetaValues is None:
            betaValues = self.betaInitValues
        else:
            if not isinstance(theBetaValues, dict):
                err = (
                    'Deprecated. A dictionary must be provided. '
                    'It can be obtained from results.getBetaValues()'
                )
                raise excep.biogemeError(err)
            for x in theBetaValues.keys():
                if x not in self.freeBetaNames:
                    self.logger.warning(f'Parameter {x} not present in the model')
            betaValues = list()
            for i, x in enumerate(self.freeBetaNames):
                if x in theBetaValues:
                    betaValues.append(theBetaValues[x])
                else:
                    self.logger.warning(
                        f'Simulation: initial value of {x} not provided.'
                    )
                    betaValues.append(self.betaInitValues[i])

        output = pd.DataFrame(index=self.database.data.index)
        for k, v in self.formulas.items():
            self.logger.detailed(f'Simulate {k}')
            signature = v.getSignature()
            result = self.theC.simulateFormula(
                signature, betaValues, self.fixedBetaValues, self.database.data
            )
            output[k] = result
        return output

    def confidenceIntervals(self, betaValues, intervalSize=0.9):
        """Calculate confidence intervals on the simulated quantities


        :param betaValues: array of parameters values to be used in
               the calculations. Typically, it is a sample drawn from
               a distribution.
        :type betaValues: list(dict(str: float))

        :param intervalSize: size of the reported confidence interval,
                    in percentage. If it is denoted by s, the interval
                    is calculated for the quantiles (1-s)/2 and
                    (1+s)/2. The default (0.9) corresponds to
                    quantiles for the confidence interval [0.05, 0.95].
        :type intervalSize: float

        :return: two pandas data frames 'left' and 'right' with the
            same dimensions. Each row corresponds to a row in the
            database, and each column to a formula. 'left' contains the
            left value of the confidence interval, and 'right' the right
            value

            Example::

                # Read the estimation results from a file
                results = res.bioResults(pickleFile = 'myModel.pickle')
                # Retrieve the names of the betas parameters that have been
                # estimated
                betas = biogeme.freeBetaNames

                # Draw 100 realization of the distribution of the estimators
                b = results.getBetasForSensitivityAnalysis(betas, size = 100)

                # Simulate the formulas using the nominal values
                simulatedValues = biogeme.simulate(betaValues)

                # Calculate the confidence intervals for each formula
                left, right = biogeme.confidenceIntervals(b, 0.9)

        :rtype: tuple of two Pandas dataframes.

        """
        listOfResults = []
        for b in betaValues:
            r = self.simulate(b)
            listOfResults += [r]
        allResults = pd.concat(listOfResults)
        r = (1.0 - intervalSize) / 2.0
        left = allResults.groupby(level=0).quantile(r)
        right = allResults.groupby(level=0).quantile(1.0 - r)
        return left, right

    def createLogFile(self, verbosity=3):
        """Creates a log file with the messages produced by Biogeme.

        The name of the file is the name of the model with an extension .log

        :param verbosity: types of messages to be captured

            - 0: no output
            - 1: warnings
            - 2: only general information
            - 3: more verbose
            - 4: debug messages

            Default: 3.

        :type verbosity: int

        """
        self.logger.createLog(fileLevel=verbosity, fileName=self.modelName)

    def __str__(self):
        r = f'{self.modelName}: database [{self.database.name}]'
        r += str(self.formulas)
        print(r)
        return r


class negLikelihood(functionToMinimize):
    """Provides the value of the function to be minimized, as well as its
    derivatives. To be used by the opimization package.

    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, like, like_deriv, scaled):
        """Constructor"""
        self.recalculate = True
        """True if the log likelihood must be recalculated
        """

        self.x = None #: Vector of unknown parameters values

        self.batch = None
        """Value betwen 0 and 1 defining the size of the batch, that is the
        percentage of the data that should be used to approximate the
        log likelihood.
        """

        self.fv = None #: value of the function

        self.gv = None #: vector with the gradient

        self.hv = None #: second derivatives matrix

        self.bhhhv = None #: BHHH matrix

        self.like = like #: function calculating the log likelihood

        self.like_deriv = like_deriv
        """function calculating the log likelihood and its derivatives.
        """

        self.scaled = scaled
        """if True, the value of the log likelihood is divided by the number
        of observations used to calculate it. In this case, the values
        with different sample sizes are comparable.
        """

    def setVariables(self, x):
        self.recalculate = True
        self.x = x
        self.fv = None
        self.gv = None
        self.hv = None
        self.bhhhv = None

    def f(self, batch=None):
        if self.x is None:
            raise excep.biogemeError('The variables must be set first.')

        if batch is not None or self.batch is not None:
            self.batch = batch
            self.recalculate = True

        if self.fv is None:
            self.recalculate = True

        if self.recalculate:
            self.fv = self.like(self.x, self.scaled, self.batch)
            self.gv = None
            self.hv = None
            self.bhhhv = None

        return -self.fv

    def f_g(self, batch=None):
        if self.x is None:
            raise excep.biogemeError('The variables must be set first.')

        if batch is not None or self.batch is not None:
            self.batch = batch
            self.recalculate = True

        if self.fv is None or self.gv is None:
            self.recalculate = True

        if self.recalculate:
            self.fv, self.gv, *_ = self.like_deriv(
                self.x, self.scaled, hessian=False, bhhh=False, batch=batch
            )
            self.hv = None
            self.bhhhv = None

        return -self.fv, -self.gv

    def f_g_h(self, batch=None):
        if self.x is None:
            raise excep.biogemeError('The variables must be set first.')

        if batch is not None or self.batch is not None:
            self.batch = batch
            self.recalculate = True

        if self.fv is None or self.gv is None or self.hv is None:
            self.recalculate = True

        if self.recalculate:
            self.fv, self.gv, self.hv, _ = self.like_deriv(
                self.x, self.scaled, hessian=True, bhhh=False, batch=batch
            )
            self.bhhhv = None

        return -self.fv, -self.gv, -self.hv

    def f_g_bhhh(self, batch=None):
        if batch is not None or self.batch is not None:
            self.batch = batch
            self.recalculate = True

        if self.x is None:
            raise excep.biogemeError('The variables must be set first.')

        if self.fv is None or self.gv is None or self.bhhhv is None:
            self.recalculate = True

        if self.recalculate:
            self.fv, self.gv, _, self.bhhhv = self.like_deriv(
                self.x, self.scaled, hessian=False, bhhh=True, batch=batch
            )
            self.hv = None

        return (-self.fv, -self.gv, -self.bhhhv)



"""File 03nestedElasticities.py

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 10:28:11 2019

 We use a previously estimated nested logit model.
 Three alternatives: public transporation, car and slow modes.
 RP data.
 We calculate disaggregate and aggregate direct point elasticities.
"""

import sys
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
import biogeme.results as res
from biogeme.expressions import Beta, Derive


# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

# Normalize the weights
sumWeight = database.data['Weight'].sum()
numberOfRows = database.data.shape[0]
normalizedWeight = Weight * numberOfRows / sumWeight

# Calculate the number of accurences of a value in the database
numberOfMales = database.count('Gender', 1)
print(f'Number of males:   {numberOfMales}')

numberOfFemales = database.count('Gender', 2)
print(f'Number of females: {numberOfFemales}')

# For more complex conditions, we use Pandas
unreportedGender = database.data[
    (database.data['Gender'] != 1) & (database.data['Gender'] != 2)
].count()['Gender']
print(f'Unreported gender: {unreportedGender}')

# List of parameters. Their value will be set later.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0, None, None, 0)
BETA_TIME_FULLTIME = Beta('BETA_TIME_FULLTIME', 0, None, None, 0)
BETA_TIME_OTHER = Beta('BETA_TIME_OTHER', 0, None, None, 0)
BETA_DIST_MALE = Beta('BETA_DIST_MALE', 0, None, None, 0)
BETA_DIST_FEMALE = Beta('BETA_DIST_FEMALE', 0, None, None, 0)
BETA_DIST_UNREPORTED = Beta('BETA_DIST_UNREPORTED', 0, None, None, 0)
BETA_COST = Beta('BETA_COST', 0, None, None, 0)

# Define new variables. Must be consistent with estimation results.
TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
MarginalCostPT_scaled = MarginalCostPT / 10
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5
male = Gender == 1
female = Gender == 2
unreportedGender = Gender == -1
fulltime = OccupStat == 1
notfulltime = OccupStat != 1

# Definition of utility functions:
V_PT = (
    ASC_PT
    + BETA_TIME_FULLTIME * TimePT_scaled * fulltime
    + BETA_TIME_OTHER * TimePT_scaled * notfulltime
    + BETA_COST * MarginalCostPT_scaled
)
V_CAR = (
    ASC_CAR
    + BETA_TIME_FULLTIME * TimeCar_scaled * fulltime
    + BETA_TIME_OTHER * TimeCar_scaled * notfulltime
    + BETA_COST * CostCarCHF_scaled
)
V_SM = (
    ASC_SM
    + BETA_DIST_MALE * distance_km_scaled * male
    + BETA_DIST_FEMALE * distance_km_scaled * female
    + BETA_DIST_UNREPORTED * distance_km_scaled * unreportedGender
)

# Associate utility functions with the numbering of alternatives
V = {0: V_PT, 1: V_CAR, 2: V_SM}

# Definition of the nests:
# 1: nests parameter
# 2: list of alternatives

MU_NOCAR = Beta('MU_NOCAR', 1.0, 1.0, None, 0)

CAR_NEST = 1.0, [1]
NO_CAR_NEST = MU_NOCAR, [0, 2]
nests = CAR_NEST, NO_CAR_NEST

# The choice model is a nested logit
prob_pt = models.nested(V, None, nests, 0)
prob_car = models.nested(V, None, nests, 1)
prob_sm = models.nested(V, None, nests, 2)

# Calculation of the direct elasticities.
# We use the 'Derive' operator to calculate the derivatives.
direct_elas_pt_time = Derive(prob_pt, 'TimePT') * TimePT / prob_pt
direct_elas_pt_cost = (
    Derive(prob_pt, 'MarginalCostPT') * MarginalCostPT / prob_pt
)
direct_elas_car_time = Derive(prob_car, 'TimeCar') * TimeCar / prob_car
direct_elas_car_cost = Derive(prob_car, 'CostCarCHF') * CostCarCHF / prob_car
direct_elas_sm_dist = Derive(prob_sm, 'distance_km') * distance_km / prob_sm

simulate = {
    'weight': normalizedWeight,
    'Prob. car': prob_car,
    'Prob. public transportation': prob_pt,
    'Prob. slow modes': prob_sm,
    'direct_elas_pt_time': direct_elas_pt_time,
    'direct_elas_pt_cost': direct_elas_pt_cost,
    'direct_elas_car_time': direct_elas_car_time,
    'direct_elas_car_cost': direct_elas_car_cost,
    'direct_elas_sm_dist': direct_elas_sm_dist,
}

biogeme = bio.BIOGEME(database, simulate)
biogeme.modelName = '03nestedElasticties'

# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='01nestedEstimation.pickle')
except FileNotFoundError:
    sys.exit(
        'Run first the script 01nestedEstimation.py in order to generate '
        'the file 01nestedEstimation.pickle.'
    )

# simulatedValues is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulatedValues = biogeme.simulate(results.getBetaValues())

# We calculate the elasticities
simulatedValues['Weighted prob. car'] = (
    simulatedValues['weight'] * simulatedValues['Prob. car']
)
simulatedValues['Weighted prob. PT'] = (
    simulatedValues['weight'] * simulatedValues['Prob. public transportation']
)
simulatedValues['Weighted prob. SM'] = (
    simulatedValues['weight'] * simulatedValues['Prob. slow modes']
)

denominator_car = simulatedValues['Weighted prob. car'].sum()
denominator_pt = simulatedValues['Weighted prob. PT'].sum()
denominator_sm = simulatedValues['Weighted prob. SM'].sum()

direct_elas_term_car_time = (
    simulatedValues['Weighted prob. car']
    * simulatedValues['direct_elas_car_time']
    / denominator_car
).sum()
print(
    f'Aggregate direct point elasticity of car wrt time: '
    f'{direct_elas_term_car_time:.3g}'
)

direct_elas_term_car_cost = (
    simulatedValues['Weighted prob. car']
    * simulatedValues['direct_elas_car_cost']
    / denominator_car
).sum()
print(
    f'Aggregate direct point elasticity of car wrt cost: '
    f'{direct_elas_term_car_cost:.3g}'
)

direct_elas_term_pt_time = (
    simulatedValues['Weighted prob. PT']
    * simulatedValues['direct_elas_pt_time']
    / denominator_pt
).sum()
print(
    f'Aggregate direct point elasticity of PT wrt time: '
    f'{direct_elas_term_pt_time:.3g}'
)

direct_elas_term_pt_cost = (
    simulatedValues['Weighted prob. PT']
    * simulatedValues['direct_elas_pt_cost']
    / denominator_pt
).sum()
print(
    f'Aggregate direct point elasticity of PT wrt cost: '
    f'{direct_elas_term_pt_cost:.3g}'
)

direct_elas_term_sm_dist = (
    simulatedValues['Weighted prob. SM']
    * simulatedValues['direct_elas_sm_dist']
    / denominator_sm
).sum()
print(
    f'Aggregate direct point elasticity of SM wrt distance: '
    f'{direct_elas_term_sm_dist:.3g}'
)



# distutils: language=c++
# cython: embedsignature=True

cimport numpy as np
import numpy as np
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp cimport bool as bool_t

ctypedef vector[unsigned long] uint_vector
ctypedef vector[uint_vector] uint_matrix
ctypedef vector[double] double_vector
ctypedef vector[double_vector] double_matrix
ctypedef vector[double_matrix] double_tensor
ctypedef vector[string] string_vector

ctypedef int[::1] uint_vector_view
ctypedef int[:, ::1] uint_matrix_view
ctypedef double[::1] double_vector_view
ctypedef double[:, ::1] double_matrix_view


cdef extern from "biogeme.h":

	cdef cppclass biogeme:
		biogeme() except +

		double calculateLikelihood(double_vector betas, 
			double_vector fixedBetas) except +

		double calculateLikeAndDerivatives(double_vector betas,
			double_vector fixedBetas,
			uint_vector betaIds,
			double* g,
			double* h,
			double* bhhh,
			bool_t hessian,
			bool_t bhhh) except +

		void setPanel(bool_t p)

		void setBounds(double_vector lb, double_vector ub)

		void simulateFormula(vector[string] loglikeSignatures,
				     double_vector betas, 
				     double_vector fixedBetas,
				     double_matrix& data,
				     double* results) except +

		void simulateSeveralFormulas(vector[string_vector] loglikeSignatures,
				     double_vector betas, 
				     double_vector fixedBetas,
				     unsigned long numberOfThreads,
				     double_matrix& data,
				     double* results) except +

		void setExpressions(vector[string] loglikeSignatures, 
						vector[string] weightSignatures,
						unsigned long numberOfThreads)

		void setData(double_matrix& d)

		void setDataMap(uint_matrix& dm)

		void setMissingData(double md)
		
		void setDraws(double_tensor& draws)


cdef class pyBiogeme:
	cdef biogeme theBiogeme

	def __cinit__(self):
		self.theBiogeme = biogeme()


	def setPanel(self,panel=True):
		self.theBiogeme.setPanel(panel)

	def calculateLikelihoodAndDerivatives(self,
	                                      betas,
					      fixedBetas,
					      betaIds,
					      gmem,
					      hmem,
					      bmem,
					      hessian,
					      bhhh,
					      draws=None):
		n = len(betas)

		if not gmem.flags['C_CONTIGUOUS']:
			print('gmem not contiguous')
			gmem = np.ascontiguousarray(gmem)
		if not hmem.flags['C_CONTIGUOUS']:
			print('hmem not contiguous')
			hmem = np.ascontiguousarray(hmem)
		if not bmem.flags['C_CONTIGUOUS']:
			print('bmem not contiguous')
			bmem = np.ascontiguousarray(bmem)

		cdef double_vector_view gmem_view = gmem
		cdef double_matrix_view hmem_view = hmem
		cdef double_matrix_view bmem_view = bmem


		f = self.theBiogeme.calculateLikeAndDerivatives(betas,
			                                        fixedBetas,
								betaIds,
								&gmem_view[0],
								&hmem_view[0,0],
								&bmem_view[0,0],
								hessian,
								bhhh)
		return f, gmem, hmem, bmem

	def setBounds(self,lb,ub):
		self.theBiogeme.setBounds(lb,ub)

	def calculateLikelihood(self, betas,fixedBetas):
		r = self.theBiogeme.calculateLikelihood(betas, fixedBetas)
		return r

	def simulateFormula(self, formula, betas, fixedBetas, d):
		n = d.shape[0]	
		r = np.empty(n)
		if not r.flags['C_CONTIGUOUS']:
			print('r not contiguous')
			r = np.ascontiguousarray(r)
		d = np.ascontiguousarray(d)

		cdef double_vector_view r_view = r
		self.theBiogeme.simulateFormula(formula,
					        betas, 
   						fixedBetas, 
						d, 
						&r_view[0])
		return r
	
	def simulateSeveralFormulas(self, formulas, betas, fixedBetas, d, nThreads):
		n = d.shape[0]
		nf = len(formulas)
		r = np.zeros([nf, n])
		if not r.flags['C_CONTIGUOUS']:
			print('r not contiguous')
			r = np.ascontiguousarray(r)
		d = np.ascontiguousarray(d)

		cdef double_matrix_view r_view = r
		self.theBiogeme.simulateSeveralFormulas(formulas,
					        	      betas, 
   						      fixedBetas, 
 						      nThreads,
						      d, 
						      &r_view[0,0])
		return r
	
	def setExpressions(self,loglikeFormulas,nbrOfThreads,weightFormulas=None):
		cdef vector[string] w
		if (weightFormulas is not None):
			w = weightFormulas
		self.theBiogeme.setExpressions(loglikeFormulas,w,nbrOfThreads)

	def setData(self, d):
		d = np.ascontiguousarray(d)
		self.theBiogeme.setData(d)

	def setDataMap(self, m):
		m = np.ascontiguousarray(m)
		self.theBiogeme.setDataMap(m)

	def setMissingData(self, md):
		self.theBiogeme.setMissingData(md)


	def setDraws(self, draws):
		draws = np.ascontiguousarray(draws)
		self.theBiogeme.setDraws(draws)

				


				





"""File: 02simpleIntegral.py

 Author: Michel Bierlaire, EPFL
 Date: Wed Dec 11 16:57:51 2019

Calculation of a simple integral using numerical integration and
Monte-Carlo integration with various types of draws, including Halton
draws base 13. It illustrates how to use draws that are not directly
available in Biogeme.

"""

# pylint: disable=invalid-name, undefined-variable

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import draws
from biogeme.expressions import exp, bioDraws, MonteCarlo


# We create a fake database with one entry, as it is required
# to store the draws
pandas = pd.DataFrame()
pandas['FakeColumn'] = [1.0]
database = db.Database('fakeDatabase', pandas)


def halton13(sampleSize, numberOfDraws):
    """
    The user can define new draws. For example, Halton draws
    with base 13, skipping the first 10 draws.
    """
    return draws.getHaltonDraws(sampleSize, numberOfDraws, base=13, skip=10)


mydraws = {'HALTON13': (halton13, 'Halton draws, base 13, skipping 10')}
database.setRandomNumberGenerators(mydraws)

integrand = exp(bioDraws('U', 'UNIFORM'))
simulatedI = MonteCarlo(integrand)

integrand_halton = exp(bioDraws('U_halton', 'UNIFORM_HALTON2'))
simulatedI_halton = MonteCarlo(integrand_halton)

integrand_halton13 = exp(bioDraws('U_halton13', 'HALTON13'))
simulatedI_halton13 = MonteCarlo(integrand_halton13)

integrand_mlhs = exp(bioDraws('U_mlhs', 'UNIFORM_MLHS'))
simulatedI_mlhs = MonteCarlo(integrand_mlhs)

trueI = exp(1.0) - 1.0

R = 20000

sampleVariance = MonteCarlo(integrand * integrand) - simulatedI * simulatedI
stderr = (sampleVariance / R) ** 0.5
error = simulatedI - trueI

sampleVariance_halton = (
    MonteCarlo(integrand_halton * integrand_halton)
    - simulatedI_halton * simulatedI_halton
)
stderr_halton = (sampleVariance_halton / R) ** 0.5
error_halton = simulatedI_halton - trueI

sampleVariance_halton13 = (
    MonteCarlo(integrand_halton13 * integrand_halton13)
    - simulatedI_halton13 * simulatedI_halton13
)
stderr_halton13 = (sampleVariance_halton13 / R) ** 0.5
error_halton13 = simulatedI_halton13 - trueI

sampleVariance_mlhs = (
    MonteCarlo(integrand_mlhs * integrand_mlhs)
    - simulatedI_mlhs * simulatedI_mlhs
)
stderr_mlhs = (sampleVariance_mlhs / R) ** 0.5
error_mlhs = simulatedI_mlhs - trueI


simulate = {
    'Analytical Integral': trueI,
    'Simulated Integral': simulatedI,
    'Sample variance   ': sampleVariance,
    'Std Error         ': stderr,
    'Error             ': error,
    'Simulated Integral (Halton)': simulatedI_halton,
    'Sample variance (Halton)   ': sampleVariance_halton,
    'Std Error (Halton)         ': stderr_halton,
    'Error (Halton)             ': error_halton,
    'Simulated Integral (Halton13)': simulatedI_halton13,
    'Sample variance (Halton13)   ': sampleVariance_halton13,
    'Std Error (Halton13)         ': stderr_halton13,
    'Error (Halton13)             ': error_halton13,
    'Simulated Integral (MLHS)': simulatedI_mlhs,
    'Sample variance (MLHS)   ': sampleVariance_mlhs,
    'Std Error (MLHS)         ': stderr_mlhs,
    'Error (MLHS)             ': error_mlhs,
}

biogeme = bio.BIOGEME(database, simulate, numberOfDraws=R)
biogeme.modelName = '02simpleIntegral'
results = biogeme.simulate()
print(f'Analytical integral:{results.iloc[0]["Analytical Integral"]:.6g}')
print('\t\tUniform\t\tHalton\t\tHalton13\tMLHS')
print(
    f'Simulated\t{results.iloc[0]["Simulated Integral"]:.6g}\t'
    f'{results.iloc[0]["Simulated Integral (Halton)"]:.6g}\t'
    f'{results.iloc[0]["Simulated Integral (Halton13)"]:.6g}\t'
    f'{results.iloc[0]["Simulated Integral (MLHS)"]:.6g}'
)
print(
    f'Sample var.\t{results.iloc[0]["Sample variance   "]:.6g}\t'
    f'{results.iloc[0]["Sample variance (Halton)   "]:.6g}\t'
    f'{results.iloc[0]["Sample variance (Halton13)   "]:.6g}\t'
    f'{results.iloc[0]["Sample variance (MLHS)   "]:.6g}'
)
print(
    f'Std error\t{results.iloc[0]["Std Error         "]:.6g}\t'
    f'{results.iloc[0]["Std Error (Halton)         "]:.6g}\t'
    f'{results.iloc[0]["Std Error (Halton13)         "]:.6g}\t'
    f'{results.iloc[0]["Std Error (MLHS)         "]:.6g}'
)
print(
    f'Error\t\t{results.iloc[0]["Error             "]:.6g}\t'
    f'{results.iloc[0]["Error (Halton)             "]:.6g}\t'
    f'{results.iloc[0]["Error (Halton13)             "]:.6g}\t'
    f'{results.iloc[0]["Error (MLHS)             "]:.6g}'
)



"""File assisted.py
:author: Michel Bierlaire, EPFL
:date: Thu Sep 17 16:21:01 2020

Assisted specification for choice models
"""

# pylint: disable=too-many-lines, invalid-name, too-many-instance-attributes
# pylint: disable=too-many-arguments, too-many-locals
# pylint: disable=too-many-public-methods, too-many-branches
# pylint: disable=too-many-statements

import random
import copy
import html
import biogeme.biogeme as bio
import biogeme.messaging as msg
import biogeme.exceptions as excep
from biogeme import vns
from biogeme.expressions import Beta, bioMultSum

logger = msg.bioMessage()


class variable:
    """
    Class representing the possible specifications of a variable
    """

    def __init__(self, name, expression):
        """
        :param name: name of the variable
        :type name: str

        :param expression: Biogeme expression of the variable.
        :type expression: :class:`biogeme.expressions.Expression`
        """

        self.name = name #: name of the variable

        self.expression = expression #: Biogeme expression for the variable

        self.active = False #: True if variable is active

        self.generic = False #: True if the variable is generic.

        self.genericName = None #: Name of the generic variable

        self.nonlinearSpec = None
        """ Function with the nonlinear specifications.
        """

        self.used = False #: True if variable used in the model.

    def __str__(self):
        """
        Print the specification of the variable
        """
        if not self.active:
            return f'{self.name} [deactivated]'

        if self.nonlinearSpec is None:
            if self.generic:
                return f'{self.name} [generic]'
            return f'{self.name} [alt. spec.]'

        if self.generic:
            return (
                f'{self.name}_{self.nonlinearSpec(self.expression)[0]}'
                f' [generic]'
            )

        return (
            f'{self.name}_{self.nonlinearSpec(self.expression)[0]}'
            f' [alt. spec.]'
        )

    def makeGeneric(self, yes, name):
        """A variable can have two status: generic or alternative
        specific. This function changes the status

        :param yes: if True, status is set to "generic".
                    If False, status is set to "alt. specific".
        :type yes: bool

        :param name: name of the generic version of the variable
        :type name: str

        """
        self.generic = yes
        self.genericName = name

    def getExpression(self):
        """Returns the biogeme expression of the specification of the variable.

        :return: expression accozunting for the status and the
            nonlinear specification.
        :rtype: :class:`biogeme.expressions.Expression`

        """
        if not self.active:
            return None
        if self.nonlinearSpec is None:
            return self.expression
        return self.nonlinearSpec(self.expression)[1]


class groupOfVariables:
    """Class representing groups of variables. All variables in the group
    will have the same nonlinear spec. They can also share the same
    coefficient.

    """

    def __init__(self, name, variables, nonlinearSpecs):
        """
        Ctor

        :param name: name of the group of variables
        :type name: str

        :param variables: list of variables in the group
        :type varariables: list(variable)

        :param nonlinearSpecs: list of possible nonlinear specifications
        :type nonlinearSpecs: list(function)
        """
        self.name = name #: name of the group of variables.

        self.variables = variables #: list of variables in the group.

        self.genericForbiden = len(self.variables) <= 1
        """ True of the group cannot be made generic
        """

        self.generic = not self.genericForbiden
        """ True if the group is generic.
        """

        self.alwaysActive = False
        """ True if the group is always active.
        """

        self.active = False #: True if the group is active.

        self.nonlinearSpecs = nonlinearSpecs
        """ list of possible nonlinear specifications
        """

        self.selection = 0
        """Index of the selected non linear specification.
        """

        self.linear = True #: True if linear specification.

    def __str__(self):
        v = [s.__str__() for s in self.variables]
        if self.generic:
            return f'{self.name}: {v}  [generic][active:{self.active}]'

        return f'{self.name}: {v} [not generic][active:{self.active}]'

    def forbidGeneric(self):
        """Forbid the generic specification in the group"""
        self.genericForbiden = True
        self.generic = False

    def forceActive(self):
        """Force the variables in the group to be active."""
        logger.detailed(
            f'Group of variables {self.name} ' f'must be in the model.'
        )
        self.activate(True)
        self.alwaysActive = True

    def activate(self, yes):
        """A group of variables can have two status: activated or not. This
        function changes the status.

        :param yes: if True, activates the group. If False,
            desactivate the group.
        :type yes: bool

        """
        if self.alwaysActive:
            return

        self.active = yes
        for v in self.variables:
            v.active = yes

    def swapActivate(self):
        """Change the activation status"""
        if self.alwaysActive:
            raise excep.biogemeError(
                f'Variable {self.name} ' f'cannot be made inactive.'
            )
        return self.activate(not self.active)

    def makeGeneric(self, yes):
        """
        A group of variables can be generic or alternative specific.
        This function changes the status.

        :param yes: if True, status is set to "generic".
                    If False, status is set to "alt. specific".
        :type yes: bool

        :raise biogemeError: if the variable cannot be made generic.
        """
        if yes and self.genericForbiden:
            raise excep.biogemeError(
                f'Variable {self.name} ' f'cannot be made generic.'
            )

        self.generic = yes
        for v in self.variables:
            v.makeGeneric(yes, self.name)

    def swapGeneric(self):
        """Change the generic/alt. specific status"""
        if self.genericForbiden:
            raise excep.biogemeError(
                f'Variable {self.name} ' f'cannot be made generic.'
            )

        return self.makeGeneric(not self.generic)

    def setLinear(self, yes):
        """A group of variables can be linear or not. This function changes
        the status.

        :param yes: if True, status is set to "linear".
                    If False, status is set to "nonlinear".
        :type yes: bool

        """
        if self.nonlinearSpecs is None:
            return
        self.linear = yes
        for v in self.variables:
            if yes:
                v.nonlinearSpec = None
            else:
                v.nonlinearSpec = self.nonlinearSpecs[self.selection]

    def swapLinear(self):
        """Change the linearity status."""
        if self.nonlinearSpecs is None:
            return
        self.setLinear(not self.linear)

    def setSelection(self, sel):
        """Set the selection of the nonlinear specification.

        :param sel: index of list self.nonlinearSpecs corresponding to
            the nonlinear spec.
        :type sel: int

        :raise biogemeError: if the index is out of range.

        """
        if self.nonlinearSpecs is None:
            return

        if sel < 0 or sel >= len(self.nonlinearSpecs):
            raise excep.biogemeError(
                f'Index {sel} out of range [0, '
                f'{len(self.nonlinearSpecs)-1}]'
            )
        self.selection = sel
        if not self.linear:
            for v in self.variables:
                v.nonlinearSpec = self.nonlinearSpecs[self.selection]

    def getDecisions(self):
        """The decision is an integer representing the decisions with
        respect to the group of variables:


        :return: decision with respect to the group of variables

        - -3 if it is inactive
        - -2 if it is active, generic and linear
        - -1 if it is active, alt. specific and linear
        - index of the nonlinear specification if active, generic
          and nonlinear.
        - 100 plus the index of the nonlinear specification if active,
          alt. specific and nonlinear.

        :rtype: int
        """
        if not self.active:
            result = -3
        elif self.linear and self.generic:
            result = -2
        elif self.linear and not self.generic:
            result = -1
        elif self.generic:
            result = self.selection
        else:
            result = 100 + self.selection

        return result

    def setDecisions(self, decision):
        """
        Implement the decision, after verifying its validity

        :param decision:

        - -3 if it is inactive
        - -2 if it is active, generic and linear
        - -1 if it is active, alt. specific and linear
        - index of the nonlinear specification if active, generic
          and nonlinear.
        - 100 plus the index of the nonlinear specification if active,
          alt. specific and nonlinear.

        :type decision: int

        :raise biogemeError: if the decision is to deactive, while
            the group should always be active.
        """
        if decision == -3 and self.alwaysActive:
            error_msg = (
                f'Group of variables {self.name} cannot be desactivated'
            )
            raise excep.biogemeError(error_msg)
        self.activate(decision != -3)
        if decision == -3:
            return
        if decision == -2:
            self.makeGeneric(True)
            self.setLinear(True)
            return
        if decision == -1:
            self.makeGeneric(False)
            self.setLinear(True)
            return
        self.setLinear(False)
        if self.nonlinearSpecs is None:
            raise excep.biogemeError(
                f'No nonlinear specification has '
                f'been provided for {self.name}'
            )

        if decision < 100:
            self.makeGeneric(True)
            self.setSelection(decision)
            return
        self.makeGeneric(False)
        self.setSelection(decision - 100)
        return


class term:
    """Class representing the possible specifications of one term of the
    utility function

    """

    def __init__(self, var, aSegmentation, bounds, validity):
        """
        Ctor

        :param var: variable of the term.
        :type var: variable

        :param aSegmentation: discrete segmentation for the parameter
        :type aSegmentation: segmentation

        :param bounds: bounds on the coefficient
        :type bounds: tuple(float, float)

        :param validity: function checking the validity of the coefficient.
        :type validity: bool f(float)

        """
        self.var = var #: variable of the term

        self.segmentation = aSegmentation
        """ discrete segmentation of the parameter
        """

        self.coef_names = None
        """names of the Beta parameters involved in the specification of the
        term.
        """

        self.validity = validity
        """ function checking the validity of the coefficient.
        """

        self.bounds = bounds #: bounds on the coefficient

        if self.var is not None:
            self.var.used = True

        if self.segmentation is not None:
            self.segmentation.used = True

    def getDecisions(self):
        """The decision is a dict, where the keys are the name of the
        socioeconomic variables, and the value are a boolean mentioning if
        it is active or not.

        :return: decision of activation of the socio-eco variables for
            the segmentation.
        :rtype: dict(str: bool)

        """
        if self.segmentation is None:
            return None
        return self.segmentation.getDecisions()

    def setDecisions(self, decisions):
        """Implement the specification decisions, represented as a dict,
        where the keys are the name of the socioeconomic variables,
        and the value are a boolean mentioning if it is active or not.

        :param decisions: decision of activation of the socio-eco variables for
            the segmentation.
        :type decisions: dict(str: bool)

        """
        if self.segmentation is None:
            return
        self.segmentation.setDecisions(decisions)

    def getBeta(self, altname):
        """
        Obtain the name of the coefficient of the term

        :param altname: name of the alternative
        :type altname: str

        :return: name of the coefficient
        :rtype: str
        """
        if self.var is None:
            return f'beta_{altname}'
        if self.var.generic:
            return f'beta_{self.var.genericName}'

        return f'beta_{self.var.name}_{altname}'

    def isValid(self, altname, estimationResults):
        """
        Check the validity of the estimated coefficient

        :param altname: name of the laternative
        :type altname: str

        :param estimationResults: results of the estimation with Biogeme
        :type estimationResults: :class:`biogeme.results.bioResults`

        :return: True if the valus is valid, False otherwise.
        :rtype: bool

        :raise biogemeError: if the parameter has not be estimated.
        """
        if self.validity is None:
            return True, None
        if not self.var.active:
            return True, None
        estimatedValues = estimationResults.getBetaValues()
        betaName = self.getBeta(altname)
        status_msg = None
        for b in self.segmentation.getBetaNames(betaName):
            val = estimatedValues.get(b)
            if val is None:
                p = list(estimatedValues.keys())
                raise excep.biogemeError(
                    f'Parameter {b}'
                    f' has not been estimated. '
                    f'Estimated parameters: '
                    f'{p}'
                )

            ok = self.validity(val)
            if not ok:
                if status_msg is None:
                    status_msg = 'Invalid parameter(s):'
                status_msg += f' {b} in alternative {altname}: {val}'

        return status_msg is None, status_msg

    def getExpression(self, altname):
        """
        Build the Biogeme expression for the term

        :param altname: name of the alternative
        :type altname: str

        :return: eypression for the term
        :rtype: :class:`biogeme.expressions.Expression`
        """
        if self.var is None:
            coef_name = self.getBeta(altname)
            if self.segmentation is None:
                return Beta(coef_name, 0, self.bounds[0], self.bounds[1], 0)

            self.coef_names = [coef_name]
            return self.segmentation.getExpression(coef_name, self.bounds)

        if not self.var.active:
            self.coef_names = None
            return None

        coef_name = self.getBeta(altname)
        if self.segmentation is None:
            self.coef_names = [coef_name]
            return self.var.getExpression() * Beta(
                coef_name, 0, self.bounds[0], self.bounds[1], 0
            )

        theBeta = self.segmentation.getExpression(coef_name, self.bounds)
        self.coef_names = list(theBeta.setOfBetas())
        return self.var.getExpression() * theBeta

    def describe(self):
        """
        Provides a short description of the term.

        :return: short description.
        :rtype: str
        """
        if self.var is None:
            result = 'Cte.'
        elif not self.var.active:
            return ''
        else:
            result = str(self.var)

        if self.segmentation is not None:
            result += ' '
            result += self.segmentation.describe()
        return result


class utility:
    """Class representing the possible specifications of a utility function"""

    def __init__(self, alternativeId, name, terms):
        """
        Ctor.

        :param alternativeId: id of the alternative.
        :type alternativeId: int

        :param name: name of the alternative
        :type name: str

        :param terms: terms of the utility function.
        :type terms: list(term)

        """
        self.name = name #: name of the alternative

        self.id = alternativeId #: id of the alternative

        self.terms = terms #: list of terms in the utility function

    def getExpression(self):
        """
        Obtain the Biogeme expression for the utility function.

        :return: Biogeme expression
        :rtype: :class:`biogeme.expressions.Expression`
        """
        theTerms = [
            t.getExpression(self.name)
            for t in self.terms
            if t.getExpression(self.name) is not None
        ]
        return bioMultSum(theTerms)


class socioEconomic:
    """Class representing  socio-economic characteristic"""

    def __init__(self, name, expression, values):
        """Ctor

        :param name: name of the segmentation variable
        :type name: str

        :param expression: Biogeme expression of the variable
        :type expression: meth:`biogeme.expressions.Expression`

        :param values: dict with values that it can take as keys, and a name
            describing them as values.
        :type values: dict(int: str)

        """
        self.name = name #: name of the segmentation variable

        self.expression = expression #: Biogeme expression of the variable

        self.values = values
        """dict with values that it can take as keys, and a name
        describing them as values.
        """

        self.active = False #: True if the segmentation variable is active.

    def combine(self, existingValues):
        """Generates the possible combinations of values,
        corresponding to segments.
        """
        if existingValues is None:
            return [
                ([self.expression], [k], [v]) for k, v in self.values.items()
            ]

        combination = []
        for k, v in self.values.items():
            for triplet in existingValues:
                combination.append(
                    (
                        triplet[0] + [self.expression],
                        triplet[1] + [k],
                        triplet[2] + [v],
                    )
                )
        return combination


class segmentation:
    """Class representing the possible segmentations"""

    def __init__(self, dictOfSocioEco):
        """
        Ctor
        """
        self.dictOfSocioEco = {
            k: socioEconomic(k, v[0], v[1]) for k, v in dictOfSocioEco.items()
        }
        """dict of object of class socioEconomic characterizing the
        segmentation.
        """

        self.listOfVariables = [] #: list of variables involved

        self.alwaysActive = False #: True if it must always be active

        self.used = False #: True if used.

    def __str__(self):
        return f'{self.dictOfSocioEco}'

    def getDecisions(self):
        """The decision is a dict, where the keys are the name of the
        socioeconomic variables, and the value are a boolean mentioning
        if it is active or not.

        :return: decision of activation of the socio-eco variables for
            the segmentation.
        :rtype: dict(str: bool)

        """
        decision = {k: v.active for k, v in self.dictOfSocioEco.items()}
        return decision

    def setDecisions(self, decisions):
        """Implement the specification decisions, represented as a dict,
        where the keys are the name of the socioeconomic variables,
        and the value are a boolean mentioning if it is active or not.

        :param decisions: decision of activation of the socio-eco variables for
            the segmentation.
        :type decisions: dict(str: bool)

        :raise biogemeError: if the name of a segmentation is unknown
        """
        for k, v in decisions.items():
            try:
                self.dictOfSocioEco[k].active = v
            except KeyError as e:
                error_msg = (
                    f'Key {k} is unknown. '
                    f'Available segmentations: '
                    f'{list(self.dictOfSocioEco.keys())}'
                )
                raise excep.biogemeError(error_msg) from e

    def isActive(self):
        """Check if there are active variables for this segmentation."""
        if self.alwaysActive:
            return True
        activeVariables = sum(
            [v.active if v is not None else True for v in self.listOfVariables]
        )
        return activeVariables > 0

    def getBetaNames(self, coef_name):
        """Get the name of the parameters for all combinations."""
        combinations = None
        for v in self.dictOfSocioEco.values():
            if v.active:
                combinations = v.combine(combinations)

        if combinations is None:
            return [coef_name]

        betas = []
        for triplet in combinations:
            theCoefName = coef_name
            for _, _, name in zip(triplet[0], triplet[1], triplet[2]):
                theCoefName = f'{theCoefName}_{name}'
            betas.append(theCoefName)
        return betas

    def getExpression(self, coef_name, bounds):
        """Obtain the Biogeme expression.

        :param coef_name: name of the coefficient
        :type coef_name: str

        :param bounds: bounds on the coefficient
        :type bounds: tuple(float, float)

        :return: biogeme  expression for the segmentation
        :rtype: :class:`biogeme.expressions.bioMultSum`
        """
        combinations = None
        for v in self.dictOfSocioEco.values():
            if v.active:
                combinations = v.combine(combinations)

        if combinations is None:
            return Beta(coef_name, 0, bounds[0], bounds[1], 0)

        listOfTerms = []
        for triplet in combinations:
            theCoefName = coef_name
            listOfConditions = []
            for var, value, name in zip(triplet[0], triplet[1], triplet[2]):
                theCoefName = f'{theCoefName}_{name}'
                listOfConditions.append(var == value)
            aTerm = Beta(theCoefName, 0, bounds[0], bounds[1], 0)
            for t in listOfConditions:
                aTerm = aTerm * t
            listOfTerms.append(aTerm)
        return bioMultSum(listOfTerms)

    def describe(self):
        """Description of the segmentation

        :return: description
        :rtype: str
        """
        active = [t.name for t in self.dictOfSocioEco.values() if t.active]
        if active:
            return '<' + ', '.join(active) + '>'
        return ''


class specificationProblem(vns.problemClass):
    """Class defining the choice model specification problem"""

    def __init__(
        self,
        name,
        database,
        theVariables,
        theGroups,
        genericForbiden,
        forceActive,
        theNonlinearSpecs,
        theSegmentations,
        utilities,
        availabilities,
        choice,
        models,
    ):
        """Ctor.

        :param name: name of the problem.
        :type name: str

        :param database: data for the estimation
        :type database: :class:`biogeme.database.Database`

        :param theVariables: variables involved in the model and their names
        :type theVariables: dict(str: :class:`biogeme.expressions.Expression`)

        :param theGroups: variables in the same groups share the same
            transforms and activation status. Each group is characterized
            by its name, and is associated to a list of variables,
            identified by their name.
        :type theGroups: dict(str: list(str))

        :param genericForbiden: groups of variables that must be
            alternative specific.

        :type genericForbiden: list(str)

        :param forceActive: groups of variables that must be in the model.
        :type forceActive: list(str)

        :param theNonlinearSpecs: associates a group of variables or a
            variable with a list of possible nonlinear
            transformations. Each transformation is a function that takes
            one argument (the variable), and return a tuple with

            - the name of the nonlinear transform
            - the expression of the transform.

            Examples of such a function::

                def sqrt(x):
                    return 'sqrt', x**0.5

                def boxcox(x):
                    ell = Beta(f'lambda', 1, 0.0001, 3.0, 0)
                    return 'Box-Cox', models.boxcox(x, ell)

        :type theNonlinearSpecs: dict(str: list( fct() ))


        :param theSegmentations: a dictionary, with keys being names
            and values beeing tuples (var, segments), where

            - var is the name of the variable
            - segments is a dict with keys being the value of the variable
              characterizing a segment, and the value being the name of
              the segment.

            Example::

                {'Income': (Income, {1: '<2500',
                                     2: '2051_4000',
                                     3: '4001_6000',
                                     4: '6001_8000',
                                     5: '8001_10000',
                                     6: '>10000',
                                     -1: 'unknown'}),
                 'Gender': (Gender, {1: 'male',
                                     2: 'female',
                                    -1: 'unkown'}),

        :type theSegmentations: dict(str, tuple(biogeme.expression.Expression,
                                                dict(int, str)))

        :param utilities: specification of the utility functions. It
            is a dict where

            - the keys are the ID of the alternatives.

            - the values are a tuple containing the name of the
              alternative and the specification.

            The specification is a list of terms. A term is a tuple with
            the name of the variable, the name of the segmentation, the
            bounds on the coeffcient, and a function checking the validity
            of the corresponding parameter (typically, check its
            sign). All can be None. If they are all None, it corresponds
            to the alternative specification constant, without any
            segmentation and any assumption on the sign.

            Example::

                utility_pt = [('PT cte', 'Seg. cte', (None, None), None),
                  ('PT travel time', 'Seg. time', (None, 0), None),
                  ('PT travel cost', 'Seg. PT cost', (None, None), isNegative),
                  ('PT Waiting time', 'Seg. wait', (None, 0), None)]


                utility_car = [('Car cte', 'Seg. cte', (None, None), None),
                  ('Car travel time', 'Seg. time', (None, 0), None),
                  (
                   'Car travel cost',
                   'Seg. car cost',
                   (None, None),
                   isNegative
                  ),
                  ('Nbr of cars', 'Seg nbr cars', (None, None), None)]

                utility_sm = [(
                               'Distance',
                               'Seg. dist',
                               (None, None),
                               isNegative
                              )]

                choiceModel = {0: ('pt', utility_pt),
                               1: ('car', utility_car),
                               2: ('sm', utility_sm)}

        :type utilities: dict(int, tuple(str,
                         list(tuple(str, str, tuple(float, float),function))))

        :param availabilities: dict describing the availability of the
            alternatives.

        :type availabilities: dict(int, :class:`biogeme.expressions.Expression`)

        :param choice: expression for the observed choice
        :type choice: :class:`biogeme.expressions.Expression`

        :param models: dict of possible models. A model is a function
            that takes the utilities and the availabilities, and return
            the loglikelihood expression.

        :type models: dict(str, fct)

        :raise biogemeError: if a variable is found in two different groups.

        :raise biogemeError: if the list of groups forbiden to be
            active contains an unknown group.

        :raise biogemeError: if the list of groups forced to be active
            contains an unknown group.

        :raise biogemeError: if some variables are not in any group.

        :raise biogemeError: if a segmentation in a utility function is
            unknown.

        :raise biogemeError: if a variable in a utility function is unknown.

        :raise biogemeError: if some variables are not used.



        """

        self.archive = {}
        """Dictionary, where the keys are solutions (objects of type
        :class:`biogeme.vns.solutionClass`) and the values are the
        estimation results (objects of type
        :class:`biogeme.results.bioResults`).
        """

        self.name = name #: name of the problem.

        self.database = database
        """object of type :class:`biogeme.database.Database`, containing the
        data.
        """

        # First check if all the variables are in a group. For those
        # who are not, create a group with a single variable
        groupForVar = {k: None for k in theVariables}
        for group, listOfVars in theGroups.items():
            for x in listOfVars:
                if groupForVar[x] is None:
                    groupForVar[x] = group
                else:
                    error_msg = (
                        f'Var {x} cannot be both in group {group}'
                        f' and group {groupForVar[x]}'
                    )
                    raise excep.biogemeError(error_msg)

        for x, g in groupForVar.items():
            if g is None:
                logger.detailed(f'Variable {x} is alone in a group.')
                theGroups[x] = [x]

        self.theVariables = {
            k: variable(k, v) for k, v in theVariables.items()
        }
        """dict of variables, where the keys are the names, and the values are
        objects of class ``variable``
        """

        self.theGroups = {
            k: groupOfVariables(
                k, [self.theVariables[i] for i in v], theNonlinearSpecs.get(k)
            )
            for k, v in theGroups.items()
        }
        """dict of groups of variables, where the keys are the names, and the
        values are objects of class ``groupOfVariables``
        """

        if genericForbiden is not None:
            for k in genericForbiden:
                try:
                    self.theGroups[k].forbidGeneric()
                except KeyError as e:
                    error_msg = (
                        f'Unknown group of variables {k} '
                        f'in the list {genericForbiden}'
                    )
                    raise excep.biogemeError(error_msg) from e

        if forceActive is not None:
            for k in forceActive:
                try:
                    self.theGroups[k].forceActive()
                except KeyError as e:
                    error_msg = (
                        f'Unknown group of variables {k} '
                        f'in the list {forceActive}'
                    )
                    raise excep.biogemeError(error_msg) from e

        check = {v: False for v in theVariables}
        for k, myvars in theGroups.items():
            for v in myvars:
                check[v] = True
        for k, c in check.items():
            if not c:
                error_msg = (
                    f'Variables not in any group: '
                    f'{[k for k, c in check.items() if not c]}'
                )
                raise excep.biogemeError(error_msg)

        self.theSegmentations = {
            k: segmentation(v) for k, v in theSegmentations.items()
        }
        """dict of segmentations, where the keys are the names, and the
        values are objects of class ``segmentation``
        """

        self.utilities = utilities
        """specification of the utility functions. See
        :class:`biogeme.assisted.specificationProblem.__init__`
        """

        self.choice = choice #: expression for the observed choice

        self.availability = availabilities
        """dict describing the availability of the alternatives.
        """

        self.theAlternatives = {}
        """Dict of utility functions, where the keys are the id of the
        alternatives, and the values are objects of type
        :class:`biogeme.assisted.utility`

        """

        # self.models = [(k, v) for k, v in models.items()]

        self.models = list(models.items())
        """ List tuple (name, model).  A model is a function
        that takes the utilities and the availabilities, and return
        the loglikelihood expression.
        """

        self.selectedModel = 0 #: index of the selected model

        self.maximumNumberOfParameters = 200
        """maximum number of parameters allowed in a specification. If the
        current model has more parameters, it is declared invalid and
        rejected by the algorithm.
        """

        self.lastOperator = None #: Last operator used

        self.operators = {
            'Change segmentation': self.changeSegmentation,
            'Increase segmentation': self.increaseSegmentation,
            'Decrease segmentation': self.decreaseSegmentation,
            'Change linearity': self.changeLinearity,
            'Change variables': self.changeVariables,
            'Change generic': self.changeGeneric,
            'Change nonlinearity': self.changeNonlinearity,
            'Change model': self.changeModel,
        }
        """Dict of operators, where the keys are the names, and the values are
        the function imple,enting the operators.
        """

        self.operatorsManagement = vns.operatorsManagement(
            self.operators.keys()
        )
        """ Object of type :class:`biogeme.vns.operatorsManagement`
        """

        # Check the consistency of the input
        for k, u in self.utilities.items():
            # All segmentations must exist
            for t in u[1]:
                # t[0] name of the variable
                # t[1] name of the segmentation
                # t[2] bounds
                # t[3] function checking validity
                if (
                    t[1] is not None
                    and self.theSegmentations.get(t[1]) is None
                ):
                    raise excep.biogemeError(
                        f'Segmentation {t[1]} ' f'does not exist'
                    )
                if t[0] is not None and self.theVariables.get(t[0]) is None:
                    raise excep.biogemeError(f'Variable {t[0]} does not exist')
                self.theAlternatives[k] = utility(
                    k,
                    u[0],
                    [
                        term(
                            self.theVariables.get(t[0]),
                            self.theSegmentations.get(t[1]),
                            t[2],
                            t[3],
                        )
                        for t in u[1]
                    ],
                )
                if t[1] is not None:
                    self.theSegmentations[t[1]].listOfVariables.append(
                        self.theVariables.get(t[0])
                    )
                # If the segmentation is not associated with a
                # variable, it must always be active
                if t[2] is None:
                    self.theSegmentations[t[1]].alwaysActive = True

        self.decisions = self.getDecisions()
        """
        The decisions consist of:

        - a dict of decisions for each group of variables.
        - for each utility, a list of decisions for each term.
        - the selected model

        Type: tuple(dict(str: int), dict(str: list(dict(str: bool))), int)
        """

        unused = list()
        for v in self.theVariables.values():
            if not v.used:
                unused.append(v.name)
        if unused:
            raise excep.biogemeError(
                f'The following variables ' f'are not used: {unused}'
            )
        unused = list()
        for s in self.theSegmentations.values():
            if not s.used:
                unused.append(s.name)
        if unused:
            raise excep.biogemeError(
                f'The following variables ' f'are not used: {unused}'
            )

    def getBiogemeModel(self):
        """Build the Biogeme expressions of a given specification"""
        V = {k: v.getExpression() for k, v in self.theAlternatives.items()}
        logprob = self.models[self.selectedModel][1](
            V, self.availability, self.choice
        )
        b = bio.BIOGEME(
            self.database,
            logprob,
            suggestScales=False,
            numberOfThreads=10,
            userNotes=self.describeHtml(),
        )
        b.generateHtml = False
        b.generatePickle = False
        return b

    def getDecisions(self):
        """
        The decisions consist of:

        - a dict of decisions for each group of variables.
        - for each utility, a list of decisions for each term.
        - the selected model

        :return: all decisions
        :rtype: tuple(dict(str: int), dict(str: list(dict(str: bool))), int)
        """

        groupDecisions = {
            k: v.getDecisions() for k, v in self.theGroups.items()
        }
        termDecisions = {
            k: [t.getDecisions() for t in u.terms]
            for k, u in self.theAlternatives.items()
        }
        return groupDecisions, termDecisions, self.selectedModel

    def setDecisions(self, decisions):
        """Implement the specification decisions

        :param decisions: specification decisions
        :type decisions: tuple(dict(str: int),
                               dict(str: list(dict(str: bool))),
                               int)

        :raise biogemeError: if a group of variables is unknown.

        :raise biogemeError: if an alternative is unknown.

        :raise biogemeError: if the length of decisions is
            inconsistent with the number of terms in the utility function.

        :raise biogemeError: if the njumber of the model is out of range.

        """
        if self.decisions == decisions:
            return

        self.decisions = decisions
        groupDecisions, termDecisions, modelDecision = decisions
        for k, v in groupDecisions.items():
            g = self.theGroups.get(k)
            if g is None:
                raise excep.biogemeError(
                    f'Unknown group of variables {k}. '
                    f'Existing groups: '
                    f'{list(self.theGroups.keys())}'
                )
            g.setDecisions(v)
        for k, v in termDecisions.items():
            u = self.theAlternatives.get(k)
            if u is None:
                raise excep.biogemeError(
                    f'Unkown alternative {u}. '
                    f'Existing alternatives: '
                    f'{list(self.utilities.keys())}'
                )
            if len(u.terms) != len(v):
                raise excep.biogemeError(
                    f'Utility of {k} has {len(u.terms)}'
                    f' terms and decisions are for '
                    f'{len(v)} terms'
                )
            for t, d in zip(u.terms, v):
                t.setDecisions(d)
        if modelDecision < 0 or modelDecision >= len(self.models):
            raise excep.biogemeError(
                f'Invalid model number: {modelDecision}.'
                f' Must be between 0 and '
                f'{len(self.models)-1}'
            )
        self.selectedModel = modelDecision

    def reset(self):
        """
        Deasactivate variables from the model, that can be deactivated.
        """
        for g in self.theGroups.values():
            g.activate(False)
        for seg in self.theSegmentations.values():
            decisions = seg.getDecisions()
            decisions = {x: False for x in decisions}
            seg.setDecisions(decisions)

    def generateSolution(self, nonlinearSpecs, segmentations, model):
        """Generate a solution for the VNS algorithm

        :param nonlinearSpecs: nonlinear specifications. It is a dictionary
            where

            - the keys correspond to groups of variables,
            - the values are tuple with two entries:

                - index in the list self.nonlinearSpecs corresponding to
                  the nonlinear spec, or None if linear,
                - a boolean that is True if the coefficient is generic,
                  False otherwise.

            Example::

                nl = {'Travel time': (0, False),
                      'Travel cost': (0, False),
                      'Headway': (0, False)}

        :type nonlinearSpecs: dict(str: tuple(int, bool))

        :param segmentations: dictionary where the keys are the name
            of the segmentations, and the values are lists of
            socio-economic characteristics that must be activated.

            Example::

                sg = {'Seg. cte': ['GA'],
                      'Seg. cost': ['class', 'who'],
                      'Seg. time': ['gender'],
                      'Seg. headway': ['class']}
        :type segmentations: dict(str: list(str))

        :param model: selected model
        :type model: str

        :return: the solution that has been generated
        :rtype: class solution

        :raise biogemeError: if a variable is set ot generic. Only groups of
            variables can be made generic.

        :raise biogemeError: if a group of variables is unknown

        :raise biogemeError: if an error occurs in setting the segmentation
            decisions

        :raise biogemeError: if the model is unknown.

        """
        self.reset()
        for k, (nl, generic) in nonlinearSpecs.items():
            if k in self.theGroups:
                self.theGroups[k].activate(True)
                self.theGroups[k].makeGeneric(generic)
                if nl is None:
                    self.theGroups[k].setLinear(True)
                else:
                    self.theGroups[k].setLinear(False)
                    self.theGroups[k].setSelection(nl)
            elif k in self.theVariables:
                self.theVariables[k].active = True
                self.theVariables[k].nonlinearSpecs = nl
                if generic:
                    raise excep.biogemeError(
                        f'Only groups of variables can '
                        f'be made generic, not {k}'
                    )
            else:
                raise excep.biogemeError(f'Unknown (group of) variable(s) {k}')

        for k, seg in segmentations.items():
            decisions = self.theSegmentations[k].getDecisions()
            decisions = {x: False for x in decisions}
            decisions = {x: True for x in seg}
            try:
                self.theSegmentations[k].setDecisions(decisions)
            except excep.biogemeError as e:
                error_msg = f'Error in segmentation of {k}: {e}'
                raise excep.biogemeError(error_msg) from e
        self.selectedModel = None
        for i, pair in enumerate(self.models):
            if pair[0] == model:
                self.selectedModel = i
        if self.selectedModel is None:
            raise excep.biogemeError(f'Unknown model {model}')
        return self.getSolution()

    def getSolution(self):
        """Generate an object of the class ``solution``

        :return: the solution that has been generated
        :rtype: class solution
        """
        s = solution()
        s.decisions = self.getDecisions()
        s.description = self.describeCurrentModel()
        s.objectives = self.archive.get(s)
        return s

    def setSolution(self, aSolution):
        """Import a solution defined as an object of class ``solution``.

        :param aSolution: solution to be imported.
        :type aSolution: class solution

        :raise biogemeError: if the object has the wrong type.
        """
        if not isinstance(aSolution, vns.solutionClass):
            raise excep.biogemeError(
                f'Wrong type: {type(aSolution)} '
                f'instead of vns.solutionClass'
            )

        self.setDecisions(aSolution.decisions)

    def clone(self):
        """Clone the model, in order to generate neighbors

        :return: a clone
        :rtype: specificationProblem
        """
        c = copy.deepcopy(self)
        return c

    def describeCurrentModel(self):
        """Generates a description of the current model

        :return: model description.
        :rtype: str
        """
        result = f'{self.models[self.selectedModel][0]}'
        for v in self.theAlternatives.values():
            title = f'Alternative {v.name} [{v.id}]\n'
            result += '-' * len(title)
            result += '\n'
            result += title
            result += '-' * len(title)
            result += '\n'
            for t in v.terms:
                result += t.describe()
                result += '\n'
        return result

    def describe(self, aSolution):
        """Generates (if necessary) and returns a short description of
            the solution

        :param aSolution: solution that must be described.
        :type aSolution: class solution

        :return: description of the solution
        :rtype: str
        """
        if aSolution.description is not None:
            return aSolution.description
        solution.description = self.describeCurrentModel()
        return solution.description

    def describeHtml(self):
        """Generates a description of the model in HTML format.

        :return: description of the model in HTML format.
        :rtype: str

        """
        result = (
            f'<p><strong>Model: {self.models[self.selectedModel][0]}'
            f'</strong></p>'
        )
        for v in self.theAlternatives.values():
            title = f'<p><strong>Alternative {v.name} [{v.id}]</strong></p>'
            result += title
            for t in v.terms:
                result += '<p>'
                result += html.escape(t.describe(), quote=True)
                result += '</p>'
                result += '\n'
        return result

    def isValid(self, aSolution):
        """Evaluate  the validity of the solution.

        :param aSolution: solution to be checked
        :type aSolution: class solution

        :return: valid, why where valid is True if the solution is
            valid, and False otherwise. ``why`` contains an explanation why it
            is invalid.
        :rtype: tuple(bool, str)
        """
        if aSolution.valid is not None:
            return aSolution.valid, aSolution.causeInvalidity

        try:
            self.setSolution(aSolution)
        except excep.biogemeError as e:
            logger.warning(f'Exception raised: {e}')
            aSolution.valid = False
            aSolution.causeInvalidity = e
            return False, e

        estimationResults = self.archive.get(aSolution)
        if estimationResults is None:
            estimationResults = self.evaluate(aSolution)
            if aSolution.valid is not None:
                return aSolution.valid, aSolution.causeInvalidity

        if estimationResults is None:
            aSolution.valid = False
            e = 'Failed estimation'
            aSolution.causeInvalidity = e
            return False, e

        if estimationResults.numberOfFreeParameters() == 0:
            aSolution.valid = False
            e = 'No parameter has been estimated'
            aSolution.causeInvalidity = e
            return False, e

        error_msg = ''
        ok = True
        for u in self.theAlternatives.values():
            for t in u.terms:
                tok, tmsg = t.isValid(u.name, estimationResults)
                if not tok:
                    ok = False
                    error_msg += f', {tmsg}'
        if ok:
            aSolution.valid = True
            return True, None

        aSolution.valid = False
        aSolution.causeInvalidity = error_msg
        return False, error_msg

    def evaluate(self, aSolution):
        """Evaluate the objectives functions of the solution and store them in
        the solution object.

        :param aSolution: solution to be evaluated
        :type aSolution: solutionClass

        :return: results of the estimation
        :rtype: class :class:`biogeme.results.bioResults`
        """
        estimationResults = self.archive.get(aSolution)
        if estimationResults is not None:
            return estimationResults

        self.setSolution(aSolution)

        b = self.getBiogemeModel()
        logger.detailed(
            f'Evaluate model with ' f'{len(b.freeBetaNames)} parameters.'
        )
        if len(b.freeBetaNames) == 0:
            estimationResults = None
            self.archive[aSolution] = estimationResults
            aSolution.valid = False
            aSolution.causeInvalidity = 'Model with 0 parameters'
            return estimationResults

        if len(b.freeBetaNames) > self.maximumNumberOfParameters:
            estimationResults = None
            self.archive[aSolution] = estimationResults
            aSolution.valid = False
            aSolution.causeInvalidity = (
                f'More than {self.maximumNumberOfParameters}'
                f' parameters: {len(b.freeBetaNames)}'
            )
            return estimationResults

        algoParameters = {'proportionAnalyticalHessian': 0.0, 'maxiter': 200}
        try:
            logger.temporarySilence()
            estimationResults = b.quickEstimate(algoParameters=algoParameters)
            logger.resume()
            aSolution.objectives = [
                -estimationResults.data.logLike,
                estimationResults.numberOfFreeParameters(),
            ]

        except excep.biogemeError as e:
            logger.warning(f'Exception raised: {e}')
            estimationResults = None
            aSolution.valid = False
            aSolution.causeInvalidity = (
                f'Exception raised during estimation: {e}'
            )
        self.archive[aSolution] = estimationResults
        return estimationResults

    def checkAvailability(self):
        """Check the availability of each operator.

        :return: a dictionary with the availability status of each operator
        :rtype: dict(str: bool)

        """
        av = {k: v(audit=True) != 0 for k, v in self.operators.items()}
        return av

    def generateNeighbor(self, aSolution, neighborhoodSize):
        """
        Generates a neighbor of the solution.

        :param aSolution: solution to be modified
        :type aSolution: class solution

        :param neighborhoodSize: size of the neighborhood to be applied
        :type neighborhoodSize: int

        :return: a neighbor solution, and the number of changes that have been
                 actually applied.
        :rtype: tuple(class solution, int)
        """
        self.setSolution(aSolution)
        # Identify the operators available for the current specification
        self.operatorsManagement.available = self.checkAvailability()
        # Select one operator.
        self.lastOperator = self.operatorsManagement.selectOperator()
        logger.detailed(f'Apply operator {self.lastOperator}')
        changes = self.applyOperator(self.lastOperator, neighborhoodSize)
        return self.getSolution(), changes

    def neighborRejected(self, aSolution, aNeighbor):
        """Notify that a neighbor has been rejected by the
        algorithm. Used to update the statistics on the operators.

        :param aSolution: solution modified. Not used in this implementation.
        :type aSolution: class solution

        :param aNeighbor: neighbor
        :type aNeighbor: solutionClass

        :raise biogemeError: if no operator has been used yet.
        """
        if self.lastOperator is None:
            raise excep.biogemeError('No operator has been used yet.')
        self.operatorsManagement.decreaseScore(self.lastOperator)

    def neighborAccepted(self, aSolution, aNeighbor):
        """Notify that a neighbor has been accepted by the
        algorithm. Used to update the statistics on the operators.

        :param aSolution: solution modified. Not used in this implementation.
        :type aSolution: solutionClass

        :param aNeighbor: neighbor
        :type aNeighbor: solutionClass

        :raise biogemeError: if no operator has been used yet.

        """
        if self.lastOperator is None:
            raise excep.biogemeError('No operator has been used yet.')
        self.operatorsManagement.increaseScore(self.lastOperator)

    def applyOperator(self, name, size=1):
        """Apply an operator.

        :param name: name of the operator to apply
        :type name: str

        :param size: size of the neighborhood
        :type size: int

        :return: total number of changes actually made on the solution
        :rtype: int

        :raise biogemeError: if the name of the operator is unknown.
        """
        op = self.operators.get(name)
        if op is None:
            raise excep.biogemeError(f'Unknowns operator: {name}')
        return op(size)

    def changeSegmentation(self, size=1, audit=False):
        """
        Change the interaction, while keeping the number of them
        """
        totalChanges = 0
        segments = list(self.theSegmentations.keys())
        random.shuffle(segments)
        for s in segments:
            if not self.theSegmentations[s].isActive():
                changes = 0
            else:
                d = self.theSegmentations[s].getDecisions()
                active = [k for k, v in d.items() if v]
                inactive = [k for k, v in d.items() if not v]
                changes = min([size, len(active), len(inactive)])
                totalChanges += changes
                if not audit:
                    for i in range(changes):
                        newd = d
                        newd[active[i]] = False
                        newd[inactive[i]] = True
                        self.theSegmentations[s].setDecisions(newd)
                if totalChanges == size:
                    return totalChanges
        return totalChanges

    def increaseSegmentation(self, size=1, audit=False):
        """
        Add a level of segmentation.
        """
        candidates = []
        segments = list(self.theSegmentations.keys())
        for s in segments:
            if self.theSegmentations[s].isActive():
                d = self.theSegmentations[s].getDecisions()
                inactive = [k for k, v in d.items() if not v]
                candidates += [(s, k) for k in inactive]

        changes = min(size, len(candidates))
        if not audit:
            random.shuffle(candidates)
            for i in range(changes):
                s, k = candidates[i]
                d = self.theSegmentations[s].getDecisions()
                d[k] = True
                d = self.theSegmentations[s].setDecisions(d)

        return changes

    def decreaseSegmentation(self, size=1, audit=False):
        """
        Remove a level of segmentation.
        """

        candidates = []
        segments = list(self.theSegmentations.keys())
        for s in segments:
            if self.theSegmentations[s].isActive():
                d = self.theSegmentations[s].getDecisions()
                active = [k for k, v in d.items() if v]
                candidates += [(s, k) for k in active]

        changes = min(size, len(candidates))
        if not audit:
            random.shuffle(candidates)
            for i in range(changes):
                s, k = candidates[i]
                d = self.theSegmentations[s].getDecisions()
                d[k] = False
                d = self.theSegmentations[s].setDecisions(d)

        return changes

    def changeLinearity(self, size=1, audit=False):
        """
        Make linear if non linear, and the other way around.
        """
        groups = [g for g in self.theGroups.keys() if self.theGroups[g].active]
        random.shuffle(groups)
        changes = min([size, len(groups)])
        if not audit:
            for i in range(changes):
                self.theGroups[groups[i]].swapLinear()

        return changes

    def changeVariables(self, size=1, audit=False):
        """
        Activate groups of variables
        """
        groups = [
            g
            for g in self.theGroups.keys()
            if not self.theGroups[g].alwaysActive
        ]
        random.shuffle(groups)
        changes = min([size, len(groups)])
        if not audit:
            for i in range(changes):
                self.theGroups[groups[i]].swapActivate()
        return changes

    def changeGeneric(self, size=1, audit=False):
        """
        Change generic vs alternative specific status
        """
        groups = [
            g
            for g in self.theGroups.keys()
            if self.theGroups[g].active
            and not self.theGroups[g].genericForbiden
        ]
        random.shuffle(groups)
        changes = min([size, len(groups)])
        if not audit:
            for i in range(changes):
                self.theGroups[groups[i]].swapGeneric()
        return changes

    def changeNonlinearity(self, size=1, audit=False):
        """
        Change the nature of the nonlinear specification
        """
        groups = [
            g
            for g in self.theGroups.keys()
            if self.theGroups[g].active and not self.theGroups[g].linear
        ]
        random.shuffle(groups)
        changes = min([size, len(groups)])
        if not audit:
            for i in range(changes):
                # We need to select randomly a specification different
                # from the current one
                nbrOfValues = len(self.theGroups[groups[i]].nonlinearSpecs)
                values = set(range(nbrOfValues))
                values.remove(self.theGroups[groups[i]].selection)
                sel = random.choice(list(values))
                self.theGroups[groups[i]].setSelection(sel)

        return changes

    def changeModel(self, size=1, audit=False):
        """
        Select randomly another model from the list

        :param size: not used here. Muxt be there for compliance with the
            call of operators.
        :type size: int

        :param audit: if True, returns the number of changes without
            actually implementing them.
        :type audit: bool

        :return: 0 if no other model could be found. 1 otherwise.
        :rtype: int
        """
        candidates = set(range(len(self.models)))
        candidates.remove(self.selectedModel)
        if not candidates:
            return 0
        if not audit:
            self.selectedModel = random.choice(list(candidates))
        return 1


class solution(vns.solutionClass):
    """
    Class representing one solution, that is, one model specification.
    """

    def __init__(self):
        super().__init__()

        self.objectivesNames = ['Neg. log likelihood', '#parameters']
        """ Names of the objective functions
        """

        self.objectives = None #: values of the objectives

        self.valid = None #: True if the solution is valid

        self.causeInvalidity = None
        """If the solution is invalid, contains the cause of the invalidity
        """

        self.decisions = None
        """
        The decisions consist of:

        - a dict of decisions for each group of variables.
        - for each utility, a list of decisions for each term.
        - the selected model

        Type: tuple(dict(str: int), dict(str: list(dict(str: bool))), int)
        """

        self.description = None #: description of the solution

    def __repr__(self):
        return str(self.decisions)

    def __str__(self):
        if self.description is None:
            raise excep.biogemeError(
                'Description of the solution not available'
            )
        res = self.description
        if self.objectivesNames is None:
            raise excep.biogemeError(
                'The attribute objectivesNames is not defined'
            )
        if self.objectives is None:
            raise excep.biogemeError(
                'The attribute objectives is not defined'
            )
        for t, r in zip(self.objectivesNames, self.objectives):
            res += f'\n{t}: {r}'
        return res
